import json
import logging
import posixpath
from abc import ABC, abstractmethod
from collections.abc import Generator, Iterable, Iterator, Sequence
from random import getrandbits
from typing import TYPE_CHECKING, Any, Optional, Union
from urllib.parse import urlparse

import attrs
import sqlalchemy as sa
from sqlalchemy import Table, and_, case, select
from sqlalchemy.sql import func
from sqlalchemy.sql.expression import true

from dvcx.client import Client
from dvcx.data_storage.schema import DATASET_CORE_COLUMN_NAMES
from dvcx.data_storage.serializer import Serializable
from dvcx.dataset import DatasetRecord, DatasetRow
from dvcx.node import DirTypeGroup, Entry, Node, NodeWithPath, get_path
from dvcx.sql.types import Int, SQLType
from dvcx.storage import StorageURI
from dvcx.utils import GLOB_CHARS, sql_escape_like

if TYPE_CHECKING:
    from sqlalchemy.sql.elements import ColumnElement
    from sqlalchemy.types import TypeEngine

    from dvcx.data_storage import AbstractIDGenerator, schema
    from dvcx.data_storage.db_engine import DatabaseEngine

try:
    import numpy as np

    numpy_imported = True
except ImportError:
    numpy_imported = False


logger = logging.getLogger("dvcx")

RANDOM_BITS = 63  # size of the random integer field

SELECT_BATCH_SIZE = 100_000  # number of rows to fetch at a time


class AbstractWarehouse(ABC, Serializable):
    """
    Abstract Warehouse class, to be implemented by any Database Adapters
    for a specific database system. This manages the storing, searching, and
    retrieval of datasets data, and has shared logic for all database
    systems currently in use.
    """

    #
    # Constants, Initialization, and Tables
    #

    DATASET_TABLE_PREFIX = "ds_"
    DATASET_SOURCE_TABLE_PREFIX = "src_"
    UDF_TABLE_NAME_PREFIX = "udf_"
    TMP_TABLE_NAME_PREFIX = "tmp_"

    id_generator: "AbstractIDGenerator"
    schema: "schema.Schema"
    db: "DatabaseEngine"

    def __init__(self, id_generator: "AbstractIDGenerator"):
        self.id_generator = id_generator

    def cleanup_for_tests(self):
        """Cleanup for tests."""

    def convert_type(self, val: Any, col_type: SQLType) -> Any:  # noqa: PLR0911
        """
        Tries to convert value to specific type if needed and if compatible,
        otherwise throws an ValueError.
        If value is a list or some other iterable, it tries to convert sub elements
        as well
        """
        if numpy_imported and isinstance(val, (np.ndarray, np.generic)):
            val = val.tolist()

        col_python_type = self.python_type(col_type)
        col_type_name = type(col_type).__name__
        value_type = type(val)

        exc = None
        try:
            if col_python_type == list and value_type in (list, tuple, set):
                if len(val) == 0:
                    return []
                item_pyton_type = self.python_type(col_type.item_type)
                if item_pyton_type != list:
                    if isinstance(val[0], item_pyton_type):
                        return val
                    if item_pyton_type == float and isinstance(val[0], int):
                        return [float(i) for i in val]
                return [self.convert_type(i, col_type.item_type) for i in val]
            # special use case with JSON type as we save it as string
            if col_python_type == dict or col_type_name == "JSON":
                if value_type == str:
                    return val
                if value_type in (dict, list):
                    return json.dumps(val)
                raise ValueError(
                    f"Cannot convert value {val!r} with type" f"{value_type} to JSON"
                )

            if isinstance(val, col_python_type):
                return val
            if col_python_type == float and isinstance(val, int):
                return float(val)
        except Exception as e:  # noqa: BLE001
            exc = e
        raise ValueError(
            f"Value {val!r} with type {value_type} incompatible for "
            f"column type {col_type_name}"
        ) from exc

    @abstractmethod
    def clone(self, use_new_connection: bool = False) -> "AbstractWarehouse":
        """Clones Warehouse implementation for some Storage input.
        Setting use_new_connection will always use a new database connection.
        New connections should only be used if needed due to errors with
        closed connections."""

    def close(self) -> None:
        """Closes any active database connections."""
        self.db.close()

    #
    # Query Tables
    #

    @abstractmethod
    def is_ready(self, timeout: Optional[int] = None) -> bool: ...

    def dataset_rows(self, dataset: DatasetRecord, version: Optional[int] = None):
        version = version or dataset.latest_version

        table_name = self.dataset_table_name(dataset.name, version)
        return self.schema.dataset_row_cls(
            table_name,
            self.db.engine,
            self.db.metadata,
            dataset.get_custom_column_types(version),
        )

    @property
    def dataset_row_cls(self):
        return self.schema.dataset_row_cls

    #
    # Query Execution
    #

    def dataset_select_paginated(
        self,
        query,
        limit: Optional[int] = None,
        order_by: tuple["ColumnElement[Any]", ...] = (),
        page_size: int = SELECT_BATCH_SIZE,
    ) -> Generator[DatasetRow, None, None]:
        """
        This is equivalent to `db.execute`, but for selecting rows in batches
        """
        cols = query.selected_columns
        cols_names = [c.name for c in cols]

        if not order_by:
            ordering = [
                cols.source,
                cols.parent,
                cols.name,
                cols.version,
                cols.etag,
            ]
        else:
            ordering = order_by  # type: ignore[assignment]

        # reset query order by and apply new order by id
        paginated_query = query.order_by(None).order_by(*ordering).limit(page_size)

        results = None
        offset = 0
        num_yielded = 0
        try:
            while True:
                if limit is not None:
                    limit -= num_yielded
                    if limit == 0:
                        break
                    if limit < page_size:
                        paginated_query = paginated_query.limit(None).limit(limit)

                results = self.db.execute(paginated_query.offset(offset))

                processed = False
                for row in results:
                    processed = True
                    yield DatasetRow.from_result_row(cols_names, row)
                    num_yielded += 1

                if not processed:
                    break  # no more results
                offset += page_size
        finally:
            # https://www2.sqlite.org/cvstrac/wiki?p=DatabaseIsLocked (SELECT not
            # finalized or reset) to prevent database table is locked error when an
            # exception is raised in the middle of processing the results (e.g.
            # https://github.com/iterative/dvcx/issues/924). Connections close
            # apparently is not enough in some cases, at least on sqlite
            # https://www.sqlite.org/c3ref/close.html
            if results and hasattr(results, "close"):
                results.close()

    #
    # Table Name Internal Functions
    #

    @staticmethod
    def uri_to_storage_info(uri: str) -> tuple[str, str]:
        parsed = urlparse(uri)
        name = parsed.path if parsed.scheme == "file" else parsed.netloc
        return parsed.scheme, name

    def dataset_table_name(self, dataset_name: str, version: int) -> str:
        prefix = self.DATASET_TABLE_PREFIX
        if Client.is_data_source_uri(dataset_name):
            # for datasets that are created for bucket listing we use different prefix
            prefix = self.DATASET_SOURCE_TABLE_PREFIX
        return f"{prefix}{dataset_name}_{version}"

    #
    # Datasets
    #

    @abstractmethod
    def create_dataset_rows_table(
        self,
        name: str,
        custom_columns: Sequence["sa.Column"] = (),
        if_not_exists: bool = True,
    ) -> Table:
        """Creates a dataset rows table for the given dataset name and columns"""

    def drop_dataset_rows_table(
        self,
        dataset: DatasetRecord,
        version: int,
        if_exists: bool = True,
    ) -> None:
        """Drops a dataset rows table for the given dataset name."""
        table_name = self.dataset_table_name(dataset.name, version)
        table = Table(table_name, self.db.metadata)
        self.db.drop_table(table, if_exists=if_exists)

    @abstractmethod
    def merge_dataset_rows(
        self,
        src: "DatasetRecord",
        dst: "DatasetRecord",
        src_version: int,
        dst_version: int,
    ) -> None:
        """
        Merges source dataset rows and current latest destination dataset rows
        into a new rows table created for new destination dataset version.
        Note that table for new destination version must be created upfront.
        Merge results should not contain duplicates.
        """

    @abstractmethod
    def dataset_rows_select(self, select_query: sa.sql.selectable.Select, **kwargs):
        """
        Method for fetching dataset rows from database. This is abstract since
        in some DBs we need to use special settings
        """

    @abstractmethod
    def get_dataset_sources(
        self, dataset: DatasetRecord, version: int
    ) -> list[StorageURI]: ...

    def get_dataset_rows(
        self,
        dataset: DatasetRecord,
        version: int,
        offset: Optional[int] = 0,
        limit: Optional[int] = 20,
        custom_columns=False,
        source: Optional[str] = None,
    ) -> Iterator[DatasetRow]:
        """Gets dataset rows."""
        if not self.db.has_table(self.dataset_table_name(dataset.name, version)):
            return

        # if custom_columns are to be returned, we are setting columns to None to
        # retrieve all columns from a table (default ones and custom)
        columns = list(DATASET_CORE_COLUMN_NAMES) if not custom_columns else None
        dr = self.dataset_rows(dataset, version)

        if not columns:
            # fetching all columns if specific columns are not defined
            columns = [c.name for c in dr.c]

        select_columns = [dr.c[c] for c in columns]
        column_index_mapping = {c: i for i, c in enumerate(columns)}

        query = dr.select(*select_columns)
        if source:
            query = query.where(dr.c.source == source)
        if offset:
            query = query.offset(offset)
        if limit:
            query = query.limit(limit)

        # ordering by primary key / index
        query = query.order_by("id")

        def map_row_to_dict(row):
            return {c: row[i] for c, i in column_index_mapping.items()}

        for row in self.dataset_rows_select(query):
            yield DatasetRow.from_dict(map_row_to_dict(row))

    def nodes_dataset_query(
        self,
        dataset_rows,
        column_names: Optional[Iterable[str]] = None,
        path: Optional[str] = None,
        recursive: Optional[bool] = False,
    ) -> "sa.Select":
        """
        Creates query pointing to certain bucket listing represented by dataset_rows
        The given `column_names`
        will be selected in the order they're given. `path` is a glob which
        will select files in matching directories, or if `recursive=True` is
        set then the entire tree under matching directories will be selected.
        """
        dr = dataset_rows

        def _is_glob(path: str) -> bool:
            return any(c in path for c in ["*", "?", "[", "]"])

        select_query = dr.dataset_query(*(column_names or []))
        if path is None:
            return select_query
        if recursive:
            root = False
            if not path or path == "/":
                # root of the bucket, e.g s3://bucket/ -> getting all the nodes
                # in the bucket
                root = True

            if not root and not _is_glob(path):
                # not a root and not a explicit glob, so it's pointing to some directory
                # and we are adding a proper glob syntax for it
                # e.g s3://bucket/dir1 -> s3://bucket/dir1/*
                path = path.rstrip("/") + "/*"

            if not root:
                # not a root, so running glob query
                select_query = select_query.where(self.path_expr(dr).op("GLOB")(path))
        else:
            parent = self.get_node_by_path(dr, path.lstrip("/").rstrip("/*"))
            select_query = select_query.where(dr.c.parent == parent.path)
        return select_query

    def rename_dataset_table(
        self,
        old_name: str,
        new_name: str,
        old_version: int,
        new_version: int,
    ) -> None:
        old_ds_table_name = self.dataset_table_name(old_name, old_version)
        new_ds_table_name = self.dataset_table_name(new_name, new_version)

        self.db.rename_table(old_ds_table_name, new_ds_table_name)

    def dataset_rows_count(self, dataset: DatasetRecord, version=None) -> int:
        """Returns total number of rows in a dataset"""
        dr = self.dataset_rows(dataset, version)
        query = select(sa.func.count()).select_from(dr.get_table())
        (res,) = self.db.execute(query)
        return res[0]

    def dataset_stats(
        self, dataset: DatasetRecord, version: int
    ) -> tuple[Optional[int], Optional[int]]:
        """
        Returns tuple with dataset stats: total number of rows and total dataset size.
        """
        if not (self.db.has_table(self.dataset_table_name(dataset.name, version))):
            return None, None

        dr = self.dataset_rows(dataset, version)
        table = dr.get_table()
        query = select(sa.func.count(), sa.func.sum(dr.c.size)).select_from(table)
        (res,) = self.db.execute(query)
        return res[0], res[1]

    def prepare_entries(
        self, uri: str, entries: Iterable[Entry]
    ) -> list[dict[str, Any]]:
        """
        Prepares bucket listing entry (row) for inserting into database
        """

        def _prepare_entry(entry: Entry):
            assert entry.dir_type is not None
            d = {
                "source": uri,
                "random": getrandbits(RANDOM_BITS),
                "parent_id": None,
                **attrs.asdict(entry),
            }
            return {
                f: d.get(f) for f in [c for c in DATASET_CORE_COLUMN_NAMES if c != "id"]
            }

        return [_prepare_entry(e) for e in entries]

    @abstractmethod
    def insert_rows(self, table: Table, rows: Iterable[dict[str, Any]]) -> None:
        """Does batch inserts of any kind of rows into table"""

    def insert_rows_done(self, table: Table) -> None:
        """
        Only needed for certain implementations
        to signal when rows inserts are complete.
        """

    @abstractmethod
    def insert_dataset_rows(self, df, dataset: DatasetRecord, version: int) -> int:
        """Inserts dataset rows directly into dataset table"""

    @abstractmethod
    def instr(self, source, target) -> "ColumnElement":
        """
        Return SQLAlchemy Boolean determining if a target substring is present in
        source string column
        """

    @abstractmethod
    def get_table(self, name: str) -> sa.Table:
        """
        Returns a SQLAlchemy Table object by name. If table doesn't exist, it should
        create it
        """

    @abstractmethod
    def dataset_table_export_file_names(
        self, dataset: DatasetRecord, version: int
    ) -> list[str]:
        """
        Returns list of file names that will be created when user runs dataset export
        """

    @abstractmethod
    def export_dataset_table(
        self,
        bucket_uri: str,
        dataset: DatasetRecord,
        version: int,
        client_config=None,
    ) -> list[str]:
        """
        Exports dataset table to the cloud, e.g to some s3 bucket
        """

    def python_type(self, col_type: Union["TypeEngine", "SQLType"]) -> Any:
        """Returns python type representation of some Sqlalchemy column type"""
        return col_type.python_type

    def add_node_type_where(
        self,
        query,
        type: Optional[str],
        include_subobjects: bool = True,
    ):
        if type is None:
            return query

        file_group: Sequence[int]
        if type in {"f", "file", "files"}:
            if include_subobjects:
                file_group = DirTypeGroup.SUBOBJ_FILE
            else:
                file_group = DirTypeGroup.FILE
        elif type in {"d", "dir", "directory", "directories"}:
            if include_subobjects:
                file_group = DirTypeGroup.SUBOBJ_DIR
            else:
                file_group = DirTypeGroup.DIR
        else:
            raise ValueError(f"invalid file type: {type!r}")

        c = query.selected_columns
        q = query.where(c.dir_type.in_(file_group))
        if not include_subobjects:
            q = q.where(c.vtype == "")
        return q

    def get_nodes(self, query) -> Iterator[Node]:
        """
        This gets nodes based on the provided query, and should be used sparingly,
        as it will be slow on any OLAP database systems.
        """
        return (Node(*row) for row in self.db.execute(query))

    def get_nodes_by_parent_path(
        self,
        dataset_rows,
        parent_path: str,
        type: Optional[str] = None,
    ) -> Iterator[Node]:
        """Gets nodes from database by parent path, with optional filtering"""
        dr = dataset_rows
        where_cond = (dr.c.parent == parent_path) & (dr.c.is_latest == true())
        if parent_path == "":
            # Exclude the root dir
            where_cond = where_cond & (dr.c.name != "")
        query = dr.select(*dr.default_columns).where(where_cond)
        query = self.add_node_type_where(query, type)
        query = query.order_by(
            dr.c.source, dr.c.parent, dr.c.name, dr.c.version, dr.c.etag
        )
        return self.get_nodes(query)

    def _get_nodes_by_glob_path_pattern(
        self, dataset_rows, path_list: list[str], glob_name: str
    ) -> Iterator[Node]:
        """Finds all Nodes that correspond to GLOB like path pattern."""
        dr = dataset_rows
        path_glob = "/".join([*path_list, glob_name])
        dirpath = path_glob[: -len(glob_name)]
        relpath = func.substr(self.path_expr(dr), len(dirpath) + 1)

        return self.get_nodes(
            dataset_rows.select(*dataset_rows.default_columns)
            .where(
                (self.path_expr(dr).op("GLOB")(path_glob))
                & (dr.c.is_latest == true())
                & ~self.instr(relpath, "/")
                & (self.path_expr(dr) != dirpath)
            )
            .order_by(dr.c.source, dr.c.parent, dr.c.name, dr.c.version, dr.c.etag)
        )

    def _get_node_by_path_list(
        self, dataset_rows, path_list: list[str], name: str
    ) -> Node:
        """
        Gets node that correspond some path list, e.g ["data-lakes", "dogs-and-cats"]
        """
        parent = "/".join(path_list)
        dr = dataset_rows

        q = (
            dr.select(*dr.default_columns)
            .where(
                (dr.c.parent == parent)
                & (dr.c.name == name)
                & (dr.c.is_latest == true())
            )
            .order_by(dr.c.source, dr.c.parent, dr.c.name, dr.c.version, dr.c.etag)
        )
        row = next(self.dataset_rows_select(q), None)
        if not row:
            path = f"{parent}/{name}"
            raise FileNotFoundError(f"Unable to resolve path {path!r}")
        return Node(*row)

    def _populate_nodes_by_path(self, dataset_rows, path_list: list[str]) -> list[Node]:
        """
        Puts all nodes found by path_list into the res input variable.
        Note that path can have GLOB like pattern matching which means that
        res can have multiple nodes as result.
        If there is no GLOB pattern, res should have one node as result that
        match exact path by path_list
        """
        if not path_list:
            return [self._get_node_by_path_list(dataset_rows, [], "")]
        matched_paths: list[list[str]] = [[]]
        for curr_name in path_list[:-1]:
            if set(curr_name).intersection(GLOB_CHARS):
                new_paths = []
                for path in matched_paths:
                    nodes = self._get_nodes_by_glob_path_pattern(
                        dataset_rows, path, curr_name
                    )
                    for node in nodes:
                        if node.is_container:
                            new_paths.append([*path, node.name or ""])
                matched_paths = new_paths
            else:
                for path in matched_paths:
                    path.append(curr_name)
        curr_name = path_list[-1]
        if set(curr_name).intersection(GLOB_CHARS):
            result: list[Node] = []
            for path in matched_paths:
                nodes = self._get_nodes_by_glob_path_pattern(
                    dataset_rows, path, curr_name
                )
                result.extend(nodes)
        else:
            result = [
                self._get_node_by_path_list(dataset_rows, path, curr_name)
                for path in matched_paths
            ]
        return result

    def get_node_by_path(self, dataset_rows, path: str) -> Node:
        """Gets node that corresponds to some path"""
        dr = dataset_rows
        query = dr.select(*dr.default_columns).where(
            (self.path_expr(dr) == path.strip("/")) & (dr.c.is_latest == true())
        )
        if path.endswith("/"):
            query = self.add_node_type_where(query, "dir")

        query = query.order_by(
            dr.c.source, dr.c.parent, dr.c.name, dr.c.version, dr.c.etag
        )
        row = next(self.db.execute(query), None)
        if not row:
            raise FileNotFoundError(f"Unable to resolve path {path}")
        return Node(*row)

    def expand_path(self, dataset_rows: DatasetRow, path: str) -> list[Node]:
        """Simulates Unix-like shell expansion"""
        clean_path = path.strip("/")
        path_list = clean_path.split("/") if clean_path != "" else []
        res = self._populate_nodes_by_path(dataset_rows, path_list)
        if path.endswith("/"):
            res = [node for node in res if node.dir_type in DirTypeGroup.SUBOBJ_DIR]
        return res

    def select_node_fields_by_parent_path(
        self,
        dataset_rows,
        parent_path: str,
        fields: Iterable[str],
    ) -> Iterator[tuple[Any, ...]]:
        """
        Gets latest-version file nodes from the provided parent path
        """
        dr = dataset_rows
        where_cond = (dr.c.parent == parent_path) & (dr.c.is_latest == true())
        if parent_path == "":
            # Exclude the root dir
            where_cond = where_cond & (dr.c.name != "")
        return self.db.execute(
            dr.select(*(getattr(dr.c, f) for f in fields))
            .where(where_cond)
            .order_by(dr.c.source, dr.c.parent, dr.c.name, dr.c.version, dr.c.etag)
        )

    def select_node_fields_by_parent_path_tar(
        self, dataset_rows, parent_path: str, fields: Iterable[str]
    ) -> Iterator[tuple[Any, ...]]:
        """
        Gets latest-version file nodes from the provided parent path
        """
        dr = dataset_rows
        dirpath = f"{parent_path}/"
        relpath = func.substr(self.path_expr(dr), len(dirpath) + 1)

        def field_to_expr(f):
            if f == "name":
                return relpath
            return getattr(dr.c, f)

        q = (
            select(*(field_to_expr(f) for f in fields))
            .where(
                self.path_expr(dr).like(f"{sql_escape_like(dirpath)}%"),
                ~self.instr(relpath, "/"),
                dr.c.is_latest == true(),
            )
            .order_by(dr.c.source, dr.c.parent, dr.c.name, dr.c.version, dr.c.etag)
        )
        return self.db.execute(q)

    def size(
        self, dataset_rows, node: Union[Node, dict[str, Any]], count_files: bool = False
    ) -> tuple[int, int]:
        """
        Calculates size of some node (and subtree below node).
        Returns size in bytes as int and total files as int
        """
        if isinstance(node, dict):
            is_dir = node.get("is_dir", node["dir_type"] in DirTypeGroup.SUBOBJ_DIR)
            node_size = node["size"]
            path = get_path(node["parent"], node["name"])
        else:
            is_dir = node.is_container
            node_size = node.size
            path = node.path
        if not is_dir:
            # Return node size if this is not a directory
            return node_size, 1

        sub_glob = posixpath.join(path, "*")
        dr = dataset_rows
        selections = [
            func.sum(dr.c.size),
        ]
        if count_files:
            selections.append(
                func.sum(dr.c.dir_type.in_(DirTypeGroup.FILE)),
            )
        results = next(
            self.db.execute(
                dr.select(*selections).where(
                    (self.path_expr(dr).op("GLOB")(sub_glob))
                    & (dr.c.is_latest == true())
                )
            ),
            (0, 0),
        )
        if count_files:
            return results[0] or 0, results[1] or 0
        return results[0] or 0, 0

    def path_expr(self, t):
        return case((t.c.parent == "", t.c.name), else_=t.c.parent + "/" + t.c.name)

    def _find_query(
        self,
        dataset_rows,
        node: Node,
        query,
        type: Optional[str] = None,
        conds=None,
        order_by: Optional[Union[str, list[str]]] = None,
    ) -> Any:
        if not conds:
            conds = []

        dr = dataset_rows
        path = self.path_expr(dr)

        if node.path:
            sub_glob = posixpath.join(node.path, "*")
            conds.append(path.op("GLOB")(sub_glob))
        else:
            conds.append(path != "")

        query = query.where(
            and_(
                *conds,
                dr.c.is_latest == true(),
            )
        )
        if type is not None:
            query = self.add_node_type_where(query, type)
        if order_by is not None:
            if isinstance(order_by, str):
                order_by = [order_by]
            query = query.order_by(*(getattr(dr.c, col) for col in order_by))
        return query

    def get_subtree_files(
        self,
        dataset_rows,
        node: Node,
        sort: Union[list[str], str, None] = None,
        include_subobjects: bool = True,
    ) -> Iterator[NodeWithPath]:
        """
        Returns all file nodes that are "below" some node.
        Nodes can be sorted as well.
        """
        dr = dataset_rows
        query = self._find_query(dr, node, dr.select(*dr.default_columns))
        query = self.add_node_type_where(query, "f", include_subobjects)

        if sort is not None:
            if not isinstance(sort, list):
                sort = [sort]
            query = query.order_by(*(sa.text(s) for s in sort))  # type: ignore [attr-defined]

        prefix_len = len(node.path)

        def make_node_with_path(row):
            sub_node = Node(*row)
            return NodeWithPath(
                sub_node, sub_node.path[prefix_len:].lstrip("/").split("/")
            )

        return map(make_node_with_path, self.db.execute(query))

    def find(
        self,
        dataset_rows,
        node: Node,
        fields: Iterable[str],
        type=None,
        conds=None,
        order_by=None,
    ) -> Iterator[tuple[Any, ...]]:
        """
        Finds nodes that match certain criteria and only looks for latest nodes
        under the passed node.
        """
        dr = dataset_rows
        query = self._find_query(
            dataset_rows,
            node,
            dr.select(*(getattr(dr.c, f) for f in fields)),
            type,
            conds,
            order_by=order_by,
        )
        return self.db.execute(query)

    def update_node(self, node_id: int, values: dict[str, Any]) -> None:
        # TODO used only in formats which will be deleted
        """Update entry of a specific node in the database."""

    def create_udf_table(
        self,
        name: str,
        custom_columns: Sequence["sa.Column"] = (),
    ) -> "sa.Table":
        """
        Create a temporary table for storing custom signals generated by a UDF.
        SQLite TEMPORARY tables cannot be directly used as they are process-specific,
        and UDFs are run in other processes when run in parallel.
        """
        tbl = sa.Table(
            name,
            sa.MetaData(),
            sa.Column("id", Int, primary_key=True),
            *custom_columns,
        )
        self.db.create_table(tbl, if_not_exists=True)
        return tbl

    def is_temp_table_name(self, name: str) -> bool:
        """Returns if the given table name refers to a temporary
        or no longer needed table."""
        if name.startswith(
            (self.TMP_TABLE_NAME_PREFIX, self.UDF_TABLE_NAME_PREFIX, "ds_shadow_")
        ) or name.endswith("_shadow"):
            return True
        return False

    def get_temp_table_names(self) -> list[str]:
        return [
            t
            for t in sa.inspect(self.db.engine).get_table_names()
            if self.is_temp_table_name(t)
        ]

    def cleanup_temp_tables(self, names: Iterable[str]) -> None:
        """
        Drop tables created temporarily when processing datasets.

        This should be implemented even if temporary tables are used to
        ensure that they are cleaned up as soon as they are no longer
        needed. When running the same `DatasetQuery` multiple times we
        may use the same temporary table names.
        """
        for name in names:
            self.db.drop_table(Table(name, self.db.metadata), if_exists=True)

    def subtract_query(
        self,
        source_query: sa.sql.selectable.Select,
        target_query: sa.sql.selectable.Select,
    ) -> sa.sql.selectable.Select:
        sq = source_query.alias("source_query")
        tq = target_query.alias("target_query")

        source_target_join = sa.join(
            sq,
            tq,
            (sq.c.source == tq.c.source)
            & (sq.c.parent == tq.c.parent)
            & (sq.c.name == tq.c.name),
            isouter=True,
        )

        return (
            select(*sq.c)
            .select_from(source_target_join)
            .where((tq.c.name == None) | (tq.c.name == ""))  # noqa: E711
        )

    def changed_query(
        self,
        source_query: sa.sql.selectable.Select,
        target_query: sa.sql.selectable.Select,
    ) -> sa.sql.selectable.Select:
        sq = source_query.alias("source_query")
        tq = target_query.alias("target_query")

        source_target_join = sa.join(
            sq,
            tq,
            (sq.c.source == tq.c.source)
            & (sq.c.parent == tq.c.parent)
            & (sq.c.name == tq.c.name),
        )

        return (
            select(*sq.c)
            .select_from(source_target_join)
            .where(
                (sq.c.last_modified > tq.c.last_modified)
                & (sq.c.is_latest == true())
                & (tq.c.is_latest == true())
            )
        )



import json

from transformers import pipeline

from dvcx.query import Object, udf
from dvcx.sql.types import JSON, String

try:
    from PIL import (
        Image,
        UnidentifiedImageError,
    )
except ImportError as exc:
    raise ImportError(
        "Missing dependency Pillow for computer vision:\n"
        "To install run:\n\n"
        "  pip install 'dvcx[cv]'\n"
    ) from exc


def read_image(raw):
    try:
        img = Image.open(raw)
    except UnidentifiedImageError:
        return None
    img.load()
    return img.convert("RGB")


def read_object(raw):
    return raw.read()


def read_text(raw):
    return read_object(raw).decode("utf-8")


@udf(
    params=(Object(read_image),),  # Columns consumed by the UDF.
    output={
        "model_output": JSON,
        "error": String,
    },  # Signals being returned by the UDF.
    method="image_processor",
)
class ImageHelper:
    def __init__(self, model, device, **kwargs):
        self.helper = pipeline(model=model, device=device)
        self.kwargs = kwargs

    def image_processor(self, imgs):
        result = self.helper(
            imgs,
            **self.kwargs,
        )
        return (json.dumps(result), "")


@udf(
    params=(Object(read_text),),  # Columns consumed by the UDF.
    output={
        "model_output": JSON,
        "error": String,
    },  # Signals being returned by the UDF.
    method="text_processor",
)
class TextHelper:
    def __init__(self, model, device, **kwargs):
        self.helper = pipeline(model=model, device=device)
        self.kwargs = kwargs

    def text_processor(self, text):
        result = self.helper(
            text,
            **self.kwargs,
        )
        return (json.dumps(result), "")


@udf(
    params=(Object(read_object),),  # Columns consumed by the UDF.
    output={
        "model_output": JSON,
        "error": String,
    },  # Signals being returned by the UDF.
    method="raw_processor",
)
class RawHelper:
    def __init__(self, model, device, **kwargs):
        self.helper = pipeline(model=model, device=device)
        self.kwargs = kwargs

    def raw_processor(self, obj):
        result = self.helper(
            obj,
            **self.kwargs,
        )
        return (json.dumps(result), "")



import os
from importlib import import_module
from typing import Any, Optional

from dvcx.catalog import Catalog
from dvcx.data_storage import AbstractIDGenerator, AbstractMetastore, AbstractWarehouse
from dvcx.data_storage.serializer import deserialize
from dvcx.data_storage.sqlite import SQLiteIDGenerator, SQLiteMetastore, SQLiteWarehouse
from dvcx.utils import get_envs_by_prefix

ID_GENERATOR_SERIALIZED = "DVCX__ID_GENERATOR"
ID_GENERATOR_IMPORT_PATH = "DVCX_ID_GENERATOR"
ID_GENERATOR_ARG_PREFIX = "DVCX_ID_GENERATOR_ARG_"
METASTORE_SERIALIZED = "DVCX__METASTORE"
METASTORE_IMPORT_PATH = "DVCX_METASTORE"
METASTORE_ARG_PREFIX = "DVCX_METASTORE_ARG_"
WAREHOUSE_SERIALIZED = "DVCX__WAREHOUSE"
WAREHOUSE_IMPORT_PATH = "DVCX_WAREHOUSE"
WAREHOUSE_ARG_PREFIX = "DVCX_WAREHOUSE_ARG_"
DISTRIBUTED_IMPORT_PATH = "DVCX_DISTRIBUTED"
DISTRIBUTED_ARG_PREFIX = "DVCX_DISTRIBUTED_ARG_"


def get_id_generator() -> "AbstractIDGenerator":
    id_generator_serialized = os.environ.get(ID_GENERATOR_SERIALIZED)
    if id_generator_serialized:
        id_generator_obj = deserialize(id_generator_serialized)
        if not isinstance(id_generator_obj, AbstractIDGenerator):
            raise RuntimeError(
                f"Deserialized ID generator is not an instance of AbstractIDGenerator: "
                f"{id_generator_obj}"
            )
        return id_generator_obj

    id_generator_import_path = os.environ.get(ID_GENERATOR_IMPORT_PATH)
    id_generator_arg_envs = get_envs_by_prefix(ID_GENERATOR_ARG_PREFIX)
    # Convert env variable names to keyword argument names by lowercasing them
    id_generator_args = {k.lower(): v for k, v in id_generator_arg_envs.items()}

    if id_generator_import_path:
        # ID generator paths are specified as (for example):
        # dvcx.data_storage.SQLiteIDGenerator
        if "." not in id_generator_import_path:
            raise RuntimeError(
                f"Invalid {ID_GENERATOR_IMPORT_PATH} import path:"
                f"{id_generator_import_path}"
            )
        module_name, _, class_name = id_generator_import_path.rpartition(".")
        id_generator = import_module(module_name)
        id_generator_class = getattr(id_generator, class_name)
    else:
        id_generator_class = SQLiteIDGenerator
    return id_generator_class(**id_generator_args)


def get_metastore(id_generator: Optional["AbstractIDGenerator"]) -> "AbstractMetastore":
    metastore_serialized = os.environ.get(METASTORE_SERIALIZED)
    if metastore_serialized:
        metastore_obj = deserialize(metastore_serialized)
        if not isinstance(metastore_obj, AbstractMetastore):
            raise RuntimeError(
                f"Deserialized Metastore is not an instance of AbstractMetastore: "
                f"{metastore_obj}"
            )
        return metastore_obj

    if id_generator is None:
        id_generator = get_id_generator()

    metastore_import_path = os.environ.get(METASTORE_IMPORT_PATH)
    metastore_arg_envs = get_envs_by_prefix(METASTORE_ARG_PREFIX)
    # Convert env variable names to keyword argument names by lowercasing them
    metastore_args = {k.lower(): v for k, v in metastore_arg_envs.items()}

    if metastore_import_path:
        # Metastore paths are specified as (for example):
        # dvcx.data_storage.SQLiteMetastore
        if "." not in metastore_import_path:
            raise RuntimeError(
                f"Invalid {METASTORE_IMPORT_PATH} import path: {metastore_import_path}"
            )
        module_name, _, class_name = metastore_import_path.rpartition(".")
        metastore = import_module(module_name)
        metastore_class = getattr(metastore, class_name)
    else:
        metastore_class = SQLiteMetastore
    return metastore_class(id_generator, **metastore_args)


def get_warehouse(id_generator: Optional["AbstractIDGenerator"]) -> "AbstractWarehouse":
    warehouse_serialized = os.environ.get(WAREHOUSE_SERIALIZED)
    if warehouse_serialized:
        warehouse_obj = deserialize(warehouse_serialized)
        if not isinstance(warehouse_obj, AbstractWarehouse):
            raise RuntimeError(
                f"Deserialized Warehouse is not an instance of AbstractWarehouse: "
                f"{warehouse_obj}"
            )
        return warehouse_obj

    if id_generator is None:
        id_generator = get_id_generator()

    warehouse_import_path = os.environ.get(WAREHOUSE_IMPORT_PATH)
    warehouse_arg_envs = get_envs_by_prefix(WAREHOUSE_ARG_PREFIX)
    # Convert env variable names to keyword argument names by lowercasing them
    warehouse_args = {k.lower(): v for k, v in warehouse_arg_envs.items()}

    if warehouse_import_path:
        # Warehouse paths are specified as (for example):
        # dvcx.data_storage.SQLiteWarehouse
        if "." not in warehouse_import_path:
            raise RuntimeError(
                f"Invalid {WAREHOUSE_IMPORT_PATH} import path: {warehouse_import_path}"
            )
        module_name, _, class_name = warehouse_import_path.rpartition(".")
        warehouse = import_module(module_name)
        warehouse_class = getattr(warehouse, class_name)
    else:
        warehouse_class = SQLiteWarehouse
    return warehouse_class(id_generator, **warehouse_args)


def get_distributed_class(**kwargs):
    distributed_import_path = os.environ.get(DISTRIBUTED_IMPORT_PATH)
    distributed_arg_envs = get_envs_by_prefix(DISTRIBUTED_ARG_PREFIX)
    # Convert env variable names to keyword argument names by lowercasing them
    distributed_args = {k.lower(): v for k, v in distributed_arg_envs.items()}

    if not distributed_import_path:
        raise RuntimeError(
            f"{DISTRIBUTED_IMPORT_PATH} import path is required "
            "for distributed UDF processing."
        )
    # Distributed class paths are specified as (for example):
    # module.classname
    if "." not in distributed_import_path:
        raise RuntimeError(
            f"Invalid {DISTRIBUTED_IMPORT_PATH} import path: {distributed_import_path}"
        )
    module_name, _, class_name = distributed_import_path.rpartition(".")
    distributed = import_module(module_name)
    distributed_class = getattr(distributed, class_name)
    return distributed_class(**distributed_args | kwargs)


def get_catalog(client_config: Optional[dict[str, Any]] = None) -> Catalog:
    """
    Function that creates Catalog instance with appropriate metastore
    and warehouse classes. Metastore class can be provided with env variable
    DVCX_METASTORE and if not provided, default one is used. Warehouse class
    can be provided with env variable DVCX_WAREHOUSE and if not provided,

    If classes expects some kwargs, they can be provided via env variables
    by adding them with prefix (DVCX_METASTORE_ARG_ and DVCX_WAREHOUSE_ARG_)
    and name of variable after, e.g. if it accepts team_id as kwargs
    we can provide DVCX_METASTORE_ARG_TEAM_ID=12345 env variable.
    """
    id_generator = get_id_generator()
    return Catalog(
        id_generator=id_generator,
        metastore=get_metastore(id_generator),
        warehouse=get_warehouse(id_generator),
        client_config=client_config,
    )



import pytest

from dvcx.sql import select
from dvcx.sql.selectable import values

DATA = [("a", 1.0, 100), ("b", 2.0, 200), ("c", 3.0, 300), ("d", 4.0, 400)]


@pytest.mark.parametrize(
    "query",
    [select(values(DATA)), select(values(DATA, ["letter", "float", "int"]))],
)
def test_select_values(warehouse, query):
    result = list(warehouse.db.execute(query))
    assert result == DATA


@pytest.mark.parametrize(
    "query",
    [
        select("c1", "c3").select_from(values(DATA)),
        select("letter", "int").select_from(values(DATA, ["letter", "float", "int"])),
    ],
)
def test_select_from_values(warehouse, query):
    result = list(warehouse.db.execute(query))
    assert result == [("a", 100), ("b", 200), ("c", 300), ("d", 400)]



"""
A simple data loader example.

This downloads and displays the first 5 images of the dataset.
"""

from contextlib import closing

from dvcx.catalog import get_catalog
from dvcx.error import DatasetNotFoundError
from dvcx.lib.param import Image
from dvcx.query import C, DatasetQuery

catalog = get_catalog()
try:
    ds = catalog.get_dataset("cats")
except DatasetNotFoundError:
    ds = (
        DatasetQuery(
            path="gcs://dvcx-datalakes/dogs-and-cats/",
            catalog=catalog,
        )
        .filter(C.name.glob("*cat*.jpg"))
        .save("cats")
    )


images = ds.limit(5).extract(Image(), cache=False)
with closing(images):
    for (img,) in images:
        img.show()



# In some cases it is more expedient for a UDF to process several rows
# in one invocation. In such cases the UDFs need to be written in a
# slightly different way:
#  * They need to accept a single function parameter, which will be a
#    list of tuples of UDF inputs.
#  * They need to return a list of tuples of result values, one tuple
#    per row, in the same order as the input parameters.
# To install script dependencies: pip install tabulate
from sqlalchemy import Integer
from tabulate import tabulate

from dvcx.query import C, DatasetQuery, udf


# Define the UDF:
@udf(
    ("parent", "name"),  # Columns consumed by the UDF.
    {
        "path_len": Integer
    },  # Signals being returned by the UDF, with the signal name and type.
    batch=10,
)
def name_len(names):
    return [(len(parent + name),) for (parent, name) in names]


if __name__ == "__main__":
    # Save as a new dataset
    DatasetQuery(path="gcs://dvcx-datalakes/dogs-and-cats/").filter(
        C.name.glob("*cat*")
    ).add_signals(name_len).save("cats_with_signal")

    # Output the contents of the new dataset.
    print(tabulate(DatasetQuery(name="cats_with_signal").results()[:10]))



from sqlalchemy.ext.compiler import compiles


def compiler_not_implemented(func, *spec):
    package = getattr(func, "package", None)
    if package is None:
        func_identifier = func.name
    else:
        func_identifier = f"{func.package}.{func.name}"

    @compiles(func, *spec)
    def raise_not_implemented(element, compiler, **kwargs):
        try:
            dialect_name = compiler.dialect.name
        except AttributeError:
            dialect_name = "unknown"
        raise NotImplementedError(
            f"Compiler not implemented for the SQLAlchemy function, {func_identifier},"
            f" with dialect, {dialect_name}. For information on adding dialect-specific"
            " compilers, see https://docs.sqlalchemy.org/en/14/core/compiler.html"
        )

    return raise_not_implemented



import ast
import io
import json
import logging
import math
import os
import os.path
import posixpath
import subprocess
import sys
import time
import traceback
from ast import Attribute, Call, Expr, Import, Load, Name, alias
from collections.abc import Iterable, Iterator, Mapping, Sequence
from contextlib import contextmanager, nullcontext
from copy import copy
from dataclasses import dataclass
from functools import cached_property, reduce
from random import shuffle
from threading import Thread
from typing import (
    IO,
    TYPE_CHECKING,
    Any,
    Callable,
    NamedTuple,
    NoReturn,
    Optional,
    Union,
)
from uuid import uuid4

import requests
import sqlalchemy as sa
import yaml
from attrs import asdict
from sqlalchemy import Column
from tqdm import tqdm

from dvcx.cache import DVCXCache
from dvcx.client import Client
from dvcx.config import get_remote_config, read_config
from dvcx.data_storage.schema import DATASET_CORE_COLUMN_NAMES
from dvcx.dataset import (
    DATASET_PREFIX,
    QUERY_DATASET_PREFIX,
    DatasetDependency,
    DatasetRecord,
    DatasetRow,
    DatasetStats,
    create_dataset_uri,
    parse_dataset_uri,
)
from dvcx.dataset import Status as DatasetStatus
from dvcx.error import (
    ClientError,
    DatasetInvalidVersionError,
    DatasetNotFoundError,
    DVCXError,
    PendingIndexingError,
    QueryScriptCancelError,
    QueryScriptCompileError,
    QueryScriptDatasetNotFound,
    QueryScriptRunError,
)
from dvcx.listing import Listing
from dvcx.node import DirType, Node, NodeWithPath
from dvcx.nodes_thread_pool import NodesThreadPool
from dvcx.remote.studio import StudioClient
from dvcx.sql.types import DateTime, SQLType, String
from dvcx.storage import Status, Storage, StorageStats, StorageURI
from dvcx.utils import (
    DVCXDir,
    batched,
    dvcx_paths_join,
    import_object,
    parse_params_string,
)

from .datasource import DataSource

if TYPE_CHECKING:
    from dvcx.data_storage import (
        AbstractIDGenerator,
        AbstractMetastore,
        AbstractWarehouse,
    )


logger = logging.getLogger("dvcx")

DEFAULT_DATASET_DIR = "dataset"
DATASET_FILE_SUFFIX = ".edvcx"

TTL_INT = 4 * 60 * 60

INDEX_INTERNAL_ERROR_MESSAGE = "Internal error on indexing"
DATASET_INTERNAL_ERROR_MESSAGE = "Internal error on creating dataset"
# exit code we use if last statement in query script is not instance of DatasetQuery
QUERY_SCRIPT_INVALID_LAST_STATEMENT_EXIT_CODE = 10
# exit code we use if query script was canceled
QUERY_SCRIPT_CANCELED_EXIT_CODE = 11

# dataset pull
PULL_DATASET_MAX_THREADS = 10
PULL_DATASET_CHUNK_TIMEOUT = 3600
PULL_DATASET_SLEEP_INTERVAL = 0.1  # sleep time while waiting for chunk to be available
PULL_DATASET_CHECK_STATUS_INTERVAL = 20  # interval to check export status in Studio


def _raise_remote_error(error_message: str) -> NoReturn:
    raise DVCXError(f"Error from server: {error_message}")


def noop(_: str):
    pass


@contextmanager
def print_and_capture(
    stream: "IO[str]", callback: Callable[[str], None] = noop
) -> "Iterator[list[str]]":
    lines: list[str] = []
    append = lines.append

    def loop() -> None:
        for line in iter(stream.readline, ""):
            print(line, end="")
            callback(line)
            append(line)

    thread = Thread(target=loop, daemon=True)
    thread.start()

    try:
        yield lines
    finally:
        thread.join()


class QueryResult(NamedTuple):
    dataset: Optional[DatasetRecord]
    version: Optional[int]
    output: str
    preview: Optional[list[dict]]


class DatasetRowsFetcher(NodesThreadPool):
    def __init__(
        self,
        metastore: "AbstractMetastore",
        warehouse: "AbstractWarehouse",
        remote_config: dict[str, Any],
        dataset_name: str,
        dataset_version: int,
        column_types: dict[str, Union[SQLType, type[SQLType]]],
        max_threads: int = PULL_DATASET_MAX_THREADS,
    ):
        super().__init__(max_threads)
        self._check_dependencies()
        self.metastore = metastore
        self.warehouse = warehouse
        self.dataset_name = dataset_name
        self.dataset_version = dataset_version
        self.column_types = column_types
        self.last_status_check: Optional[float] = None

        self.studio_client = StudioClient(
            remote_config["url"], remote_config["username"], remote_config["token"]
        )

    def done_task(self, done):
        for task in done:
            task.result()

    def _check_dependencies(self) -> None:
        try:
            import lz4.frame  # noqa: F401
            import numpy as np  # noqa: F401
            import pandas as pd  # noqa: F401
            import pyarrow as pa  # noqa: F401
        except ImportError as exc:
            raise Exception(
                f"Missing dependency: {exc.name}\n"
                "To install run:\n"
                "\tpip install 'dvcx[remote]'"
            ) from None

    def should_check_for_status(self) -> bool:
        if not self.last_status_check:
            return True
        return time.time() - self.last_status_check > PULL_DATASET_CHECK_STATUS_INTERVAL

    def check_for_status(self) -> None:
        """
        Method that checks export status in Studio and raises Exception if export
        failed or was removed.
        Checks are done every PULL_DATASET_CHECK_STATUS_INTERVAL seconds
        """
        export_status_response = self.studio_client.dataset_export_status(
            self.dataset_name, self.dataset_version
        )
        if not export_status_response.ok:
            _raise_remote_error(export_status_response.message)

        export_status = export_status_response.data["status"]  # type: ignore [index]

        if export_status == "failed":
            _raise_remote_error("Dataset export failed in Studio")
        if export_status == "removed":
            _raise_remote_error("Dataset export removed in Studio")

        self.last_status_check = time.time()

    def fix_columns(self, df) -> None:
        import pandas as pd

        """
        Method that does various column decoding or parsing, depending on a type
        before inserting into DB
        """
        # we get dataframe from parquet export files where datetimes are serialized
        # as timestamps so we need to parse it back to datetime objects
        for c in [c for c, t in self.column_types.items() if t == DateTime]:
            df[c] = pd.to_datetime(df[c], unit="s")

        # strings are represented as binaries in parquet export so need to
        # decode it back to strings
        for c in [c for c, t in self.column_types.items() if t == String]:
            df[c] = df[c].str.decode("utf-8")

    def do_task(self, urls):
        import lz4.frame
        import pandas as pd

        metastore = self.metastore.clone()  # metastore is not thread safe
        warehouse = self.warehouse.clone()  # warehouse is not thread safe
        dataset = metastore.get_dataset(self.dataset_name)

        urls = list(urls)
        while urls:
            for url in urls:
                if self.should_check_for_status():
                    self.check_for_status()

                r = requests.get(url, timeout=PULL_DATASET_CHUNK_TIMEOUT)
                if r.status_code == 404:
                    time.sleep(PULL_DATASET_SLEEP_INTERVAL)
                    # moving to the next url
                    continue

                r.raise_for_status()

                df = pd.read_parquet(io.BytesIO(lz4.frame.decompress(r.content)))

                self.fix_columns(df)

                # id will be autogenerated in DB
                df = df.drop("id", axis=1)

                inserted = warehouse.insert_dataset_rows(
                    df, dataset, self.dataset_version
                )
                self.increase_counter(inserted)  # type: ignore [arg-type]
                urls.remove(url)


@dataclass
class NodeGroup:
    """Class for a group of nodes from the same source"""

    listing: Listing
    sources: list[DataSource]

    # The source path within the bucket
    # (not including the bucket name or s3:// prefix)
    source_path: str = ""
    is_edvcx: bool = False
    dataset_name: Optional[str] = None
    dataset_version: Optional[int] = None
    instantiated_nodes: Optional[list[NodeWithPath]] = None

    @property
    def is_dataset(self) -> bool:
        return bool(self.dataset_name)

    def iternodes(self, recursive: bool = False):
        for src in self.sources:
            if recursive and src.is_container():
                for nwp in src.find():
                    yield nwp.n
            else:
                yield src.node

    def download(self, recursive: bool = False, pbar=None) -> None:
        """
        Download this node group to cache.
        """
        if self.sources:
            self.listing.client.fetch_nodes(
                self.iternodes(recursive), shared_progress_bar=pbar
            )


def check_output_dataset_file(
    output: str,
    force: bool = False,
    dataset_filename: Optional[str] = None,
    skip_check_edvcx: bool = False,
) -> str:
    """
    Checks the dataset filename for existence or if it should be force-overwritten.
    """
    dataset_file = (
        dataset_filename if dataset_filename else output + DATASET_FILE_SUFFIX
    )
    if not skip_check_edvcx and os.path.exists(dataset_file):
        if force:
            os.remove(dataset_file)
        else:
            raise RuntimeError(f"Output dataset file already exists: {dataset_file}")
    return dataset_file


def parse_edvcx_file(filename: str) -> list[dict[str, Any]]:
    with open(filename, encoding="utf-8") as f:
        contents = yaml.safe_load(f)

    if not isinstance(contents, list):
        contents = [contents]

    for entry in contents:
        if not isinstance(entry, dict):
            raise TypeError(
                "Failed parsing EDVCX file, "
                "each data source entry must be a dictionary"
            )
        if "data-source" not in entry or "files" not in entry:
            raise ValueError(
                "Failed parsing EDVCX file, "
                "each data source entry must contain the "
                '"data-source" and "files" keys'
            )

    return contents


def prepare_output_for_cp(
    node_groups: list[NodeGroup],
    output: str,
    force: bool = False,
    edvcx_only: bool = False,
    no_edvcx_file: bool = False,
) -> tuple[bool, Optional[str]]:
    total_node_count = 0
    for node_group in node_groups:
        if not node_group.sources:
            raise FileNotFoundError(
                f"No such file or directory: {node_group.source_path}"
            )
        total_node_count += len(node_group.sources)

    always_copy_dir_contents = False
    copy_to_filename = None

    if edvcx_only:
        return always_copy_dir_contents, copy_to_filename

    if not os.path.isdir(output):
        if all(n.is_dataset for n in node_groups):
            os.mkdir(output)
        elif total_node_count == 1:
            first_source = node_groups[0].sources[0]
            if first_source.is_container():
                if os.path.exists(output):
                    if force:
                        os.remove(output)
                    else:
                        raise FileExistsError(f"Path already exists: {output}")
                always_copy_dir_contents = True
                os.mkdir(output)
            else:  # Is a File
                if os.path.exists(output):
                    if force:
                        os.remove(output)
                    else:
                        raise FileExistsError(f"Path already exists: {output}")
                copy_to_filename = output
        else:
            raise FileNotFoundError(f"Is not a directory: {output}")

    if copy_to_filename and not no_edvcx_file:
        raise RuntimeError("File to file cp not supported with .edvcx files!")

    return always_copy_dir_contents, copy_to_filename


def collect_nodes_for_cp(
    node_groups: Iterable[NodeGroup],
    recursive: bool = False,
) -> tuple[int, int]:
    total_size: int = 0
    total_files: int = 0

    # Collect all sources to process
    for node_group in node_groups:
        listing: Listing = node_group.listing
        valid_sources: list[DataSource] = []
        for dsrc in node_group.sources:
            if dsrc.is_single_object():
                total_size += dsrc.node.size
                total_files += 1
                valid_sources.append(dsrc)
            else:
                node = dsrc.node
                if not recursive:
                    print(f"{node.full_path} is a directory (not copied).")
                    continue
                add_size, add_files = listing.du(node, count_files=True)
                total_size += add_size
                total_files += add_files
                valid_sources.append(dsrc)

        node_group.sources = valid_sources

    return total_size, total_files


def get_download_bar(bar_format: str, total_size: int):
    return tqdm(
        desc="Downloading files: ",
        unit="B",
        bar_format=bar_format,
        unit_scale=True,
        unit_divisor=1000,
        total=total_size,
    )


def instantiate_node_groups(
    node_groups: Iterable[NodeGroup],
    output: str,
    bar_format: str,
    total_files: int,
    force: bool = False,
    recursive: bool = False,
    virtual_only: bool = False,
    always_copy_dir_contents: bool = False,
    copy_to_filename: Optional[str] = None,
) -> None:
    instantiate_progress_bar = (
        None
        if virtual_only
        else tqdm(
            desc=f"Instantiating {output}: ",
            unit=" f",
            bar_format=bar_format,
            unit_scale=True,
            unit_divisor=1000,
            total=total_files,
        )
    )

    output_dir = output
    if copy_to_filename:
        output_dir = os.path.dirname(output)
        if not output_dir:
            output_dir = "."

    # Instantiate these nodes
    for node_group in node_groups:
        if not node_group.sources:
            continue
        listing: Listing = node_group.listing
        source_path: str = node_group.source_path

        copy_dir_contents = always_copy_dir_contents or source_path.endswith("/")
        instantiated_nodes = listing.collect_nodes_to_instantiate(
            node_group.sources,
            copy_to_filename,
            recursive,
            copy_dir_contents,
            source_path,
            node_group.is_edvcx,
            node_group.is_dataset,
        )
        if not virtual_only:
            listing.instantiate_nodes(
                instantiated_nodes,
                output_dir,
                total_files,
                force=force,
                shared_progress_bar=instantiate_progress_bar,
            )
        node_group.instantiated_nodes = instantiated_nodes
    if instantiate_progress_bar:
        instantiate_progress_bar.close()


def compute_metafile_data(node_groups) -> list[dict[str, Any]]:
    metafile_data = []
    for node_group in node_groups:
        if not node_group.sources:
            continue
        listing: Listing = node_group.listing
        source_path: str = node_group.source_path
        if not node_group.is_dataset:
            assert listing.storage
            data_source = listing.storage.to_dict(source_path)
        else:
            data_source = {"uri": listing.metastore.uri}

        metafile_group = {"data-source": data_source, "files": []}
        for node in node_group.instantiated_nodes:
            if not node.n.is_dir:
                metafile_group["files"].append(node.get_metafile_data())
        if metafile_group["files"]:
            metafile_data.append(metafile_group)

    return metafile_data


def find_column_to_str(  # noqa: PLR0911
    row: tuple[Any, ...], field_lookup: dict[str, int], src: DataSource, column: str
) -> str:
    if column == "du":
        return str(
            src.listing.du(
                {
                    f: row[field_lookup[f]]
                    for f in ["dir_type", "size", "parent", "name"]
                }
            )[0]
        )
    if column == "name":
        return row[field_lookup["name"]] or ""
    if column == "owner":
        return row[field_lookup["owner_name"]] or ""
    if column == "path":
        is_dir = row[field_lookup["dir_type"]] == DirType.DIR
        parent = row[field_lookup["parent"]]
        name = row[field_lookup["name"]]
        path = f"{parent}/{name}" if parent else name
        if is_dir and path:
            full_path = path + "/"
        else:
            full_path = path
        return src.get_node_full_path_from_path(full_path)
    if column == "size":
        return str(row[field_lookup["size"]])
    if column == "type":
        dt = row[field_lookup["dir_type"]]
        if dt == DirType.DIR:
            return "d"
        if dt == DirType.FILE:
            return "f"
        if dt == DirType.TAR_ARCHIVE:
            return "t"
        # Unknown - this only happens if a type was added elsewhere but not here
        return "u"
    return ""


class Catalog:
    def __init__(
        self,
        id_generator: "AbstractIDGenerator",
        metastore: "AbstractMetastore",
        warehouse: "AbstractWarehouse",
        cache_dir=None,
        tmp_dir=None,
        client_config: Optional[dict[str, Any]] = None,
        warehouse_ready_callback: Optional[
            Callable[["AbstractWarehouse"], None]
        ] = None,
    ):
        dvcx_dir = DVCXDir(cache=cache_dir, tmp=tmp_dir)
        dvcx_dir.init()
        self.id_generator = id_generator
        self.metastore = metastore
        self._warehouse = warehouse
        self.cache = DVCXCache(dvcx_dir.cache, dvcx_dir.tmp)
        self.client_config = client_config if client_config is not None else {}
        self._init_params = {
            "cache_dir": cache_dir,
            "tmp_dir": tmp_dir,
        }
        self._warehouse_ready_callback = warehouse_ready_callback

    @cached_property
    def warehouse(self) -> "AbstractWarehouse":
        if self._warehouse_ready_callback:
            self._warehouse_ready_callback(self._warehouse)

        return self._warehouse

    def get_init_params(self) -> dict[str, Any]:
        return {
            **self._init_params,
            "client_config": self.client_config,
        }

    def copy(self, cache=True, db=True):
        result = copy(self)
        if not db:
            result.id_generator = None
            result.metastore = None
            result.warehouse = None
        return result

    @classmethod
    def generate_query_dataset_name(cls) -> str:
        return f"{QUERY_DATASET_PREFIX}_{uuid4().hex}"

    def compile_query_script(self, script: str) -> str:
        code_ast = ast.parse(script)
        if code_ast.body:
            last_expr = code_ast.body[-1]
            if isinstance(last_expr, Expr):
                new_expressions = [
                    Import(names=[alias(name="dvcx.query.dataset", asname=None)]),
                    Expr(
                        value=Call(
                            func=Attribute(
                                value=Attribute(
                                    value=Attribute(
                                        value=Name(id="dvcx", ctx=Load()),
                                        attr="query",
                                        ctx=Load(),
                                    ),
                                    attr="dataset",
                                    ctx=Load(),
                                ),
                                attr="query_wrapper",
                                ctx=Load(),
                            ),
                            args=[last_expr],
                            keywords=[],
                        )
                    ),
                ]
                code_ast.body[-1:] = new_expressions
            else:
                raise Exception("Last line in a script was not an expression")

        return ast.unparse(code_ast)

    def parse_url(self, uri: str, **config: Any) -> tuple[Client, str]:
        config = config or self.client_config
        return Client.parse_url(uri, self.metastore, self.cache, **config)

    def get_client(self, uri: StorageURI, **config: Any) -> Client:
        """
        Return the client corresponding to the given source `uri`.
        """
        config = config or self.client_config
        cls = Client.get_implementation(uri)
        return cls.from_source(uri, self.cache, **config)

    def enlist_source(
        self,
        source: str,
        ttl: int,
        force_update=False,
        skip_indexing=False,
        client_config=None,
    ) -> tuple[Listing, str]:
        if force_update and skip_indexing:
            raise ValueError(
                "Both force_update and skip_indexing flags"
                " cannot be True at the same time"
            )

        partial_id: Optional[int]
        partial_path: Optional[str]

        client_config = client_config or self.client_config
        client, path = self.parse_url(source, **client_config)
        prefix = posixpath.dirname(path)
        storage_dataset_name = Storage.dataset_name(
            client.uri, posixpath.join(prefix, "")
        )
        source_metastore = self.metastore.clone(client.uri)
        source_warehouse = self.warehouse.clone()

        if skip_indexing:
            source_metastore.create_storage_if_not_registered(client.uri)
            storage = source_metastore.get_storage(client.uri)
            source_metastore.init_partial_id(client.uri)
            partial_id = source_metastore.get_next_partial_id(client.uri)

            source_metastore = self.metastore.clone(
                uri=client.uri, partial_id=partial_id
            )
            source_metastore.init(client.uri, partial_id)

            source_warehouse = self.warehouse.clone()
            dataset = self.create_dataset(storage_dataset_name, listing=True)

            return (
                Listing(storage, source_metastore, source_warehouse, client, dataset),
                path,
            )

        (
            storage,
            need_index,
            in_progress,
            partial_id,
            partial_path,
        ) = source_metastore.register_storage_for_indexing(
            client.uri, force_update, prefix
        )
        if in_progress:
            raise PendingIndexingError(f"Pending indexing operation: uri={storage.uri}")

        if not need_index:
            assert partial_id is not None
            assert partial_path is not None
            source_metastore = self.metastore.clone(
                uri=client.uri, partial_id=partial_id
            )
            source_warehouse = self.warehouse.clone()
            dataset = self.get_dataset(Storage.dataset_name(client.uri, partial_path))
            lst = Listing(storage, source_metastore, source_warehouse, client, dataset)
            logger.debug(
                "Using cached listing %s. Valid till: %s",
                storage.uri,
                storage.expires_to_local,
            )
            # Listing has to have correct version of data storage
            # initialized with correct Storage
            return lst, path

        source_metastore.init_partial_id(client.uri)
        partial_id = source_metastore.get_next_partial_id(client.uri)

        source_metastore.init(client.uri, partial_id)
        source_metastore = self.metastore.clone(uri=client.uri, partial_id=partial_id)

        source_warehouse = self.warehouse.clone()

        dataset = self.create_dataset(storage_dataset_name, listing=True)

        lst = Listing(storage, source_metastore, source_warehouse, client, dataset)

        try:
            lst.fetch(prefix)

            source_metastore.mark_storage_indexed(
                storage.uri,
                Status.PARTIAL if prefix else Status.COMPLETE,
                ttl,
                prefix=prefix,
                partial_id=partial_id,
                dataset=dataset,
            )
        except ClientError as e:
            # for handling cloud errors
            error_message = INDEX_INTERNAL_ERROR_MESSAGE
            if e.error_code in ["InvalidAccessKeyId", "SignatureDoesNotMatch"]:
                error_message = "Invalid cloud credentials"

            source_metastore.mark_storage_indexed(
                storage.uri,
                Status.FAILED,
                ttl,
                prefix=prefix,
                error_message=error_message,
                error_stack=traceback.format_exc(),
                dataset=dataset,
            )
            raise
        except:
            source_metastore.mark_storage_indexed(
                storage.uri,
                Status.FAILED,
                ttl,
                prefix=prefix,
                error_message=INDEX_INTERNAL_ERROR_MESSAGE,
                error_stack=traceback.format_exc(),
                dataset=dataset,
            )
            raise

        lst.storage = storage

        return lst, path

    def enlist_sources(
        self,
        sources: list[str],
        ttl: int,
        update: bool,
        skip_indexing=False,
        client_config=None,
        only_index=False,
    ) -> Optional[list["DataSource"]]:
        enlisted_sources = []
        for src in sources:  # Opt: parallel
            listing, file_path = self.enlist_source(
                src,
                ttl,
                update,
                skip_indexing=skip_indexing,
                client_config=client_config or self.client_config,
            )
            enlisted_sources.append((listing, file_path))

        if only_index:
            # sometimes we don't really need listing result (e.g on indexing process)
            # so this is to improve performance
            return None

        dsrc_all = []
        for listing, file_path in enlisted_sources:
            nodes = listing.expand_path(file_path)
            dir_only = file_path.endswith("/")
            for node in nodes:
                dsrc_all.append(DataSource(listing, node, dir_only))

        return dsrc_all

    def enlist_sources_grouped(
        self,
        sources: list[str],
        ttl: int,
        update: bool,
        no_glob: bool = False,
        client_config=None,
    ) -> list[NodeGroup]:
        def _ds_row_to_node(dr: DatasetRow) -> Node:
            d = asdict(dr)  # type: ignore [arg-type]
            del d["source"]
            del d["custom"]
            return Node(**d)

        enlisted_sources: list[tuple[bool, bool, Any]] = []
        client_config = client_config or self.client_config
        for src in sources:  # Opt: parallel
            if src.endswith(DATASET_FILE_SUFFIX) and os.path.isfile(src):
                # TODO: Also allow using EDVCX files from cloud locations?
                edvcx_data = parse_edvcx_file(src)
                indexed_sources = []
                for ds in edvcx_data:
                    listing, source_path = self.enlist_source(
                        ds["data-source"]["uri"],
                        ttl,
                        update,
                        client_config=client_config,
                    )
                    paths = dvcx_paths_join(
                        source_path, (f["name"] for f in ds["files"])
                    )
                    indexed_sources.append((listing, source_path, paths))
                enlisted_sources.append((True, False, indexed_sources))
            elif src.startswith("ds://"):
                ds_name, ds_version = parse_dataset_uri(src)
                dataset = self.get_dataset(ds_name)
                if not ds_version:
                    ds_version = dataset.latest_version
                dataset_sources = self.warehouse.get_dataset_sources(
                    dataset,
                    ds_version,
                )
                indexed_sources = []
                for source in dataset_sources:
                    client = self.get_client(source, **client_config)
                    uri = client.uri
                    ms = self.metastore.clone(uri, None)
                    st = self.warehouse.clone()
                    listing = Listing(None, ms, st, client, None)
                    rows = st.get_dataset_rows(
                        dataset,
                        ds_version,
                        limit=None,
                        source=source,
                    )
                    indexed_sources.append(
                        (
                            listing,
                            source,
                            rows,
                            ds_name,
                            ds_version,
                        )  # type: ignore [arg-type]
                    )

                enlisted_sources.append((False, True, indexed_sources))
            else:
                listing, source_path = self.enlist_source(
                    src, ttl, update, client_config=client_config
                )
                enlisted_sources.append((False, False, (listing, source_path)))

        node_groups = []
        for is_dvcx, is_dataset, payload in enlisted_sources:  # Opt: parallel
            if is_dataset:
                for (
                    listing,
                    source_path,
                    dataset_rows,
                    dataset_name,
                    dataset_version,
                ) in payload:
                    nodes = [_ds_row_to_node(row) for row in dataset_rows]
                    dsrc = [DataSource(listing, node) for node in nodes]
                    node_groups.append(
                        NodeGroup(
                            listing,
                            dsrc,
                            source_path,
                            dataset_name=dataset_name,
                            dataset_version=dataset_version,
                        )
                    )
            elif is_dvcx:
                for listing, source_path, paths in payload:
                    dsrc = [DataSource(listing, listing.resolve_path(p)) for p in paths]
                    node_groups.append(
                        NodeGroup(listing, dsrc, source_path, is_edvcx=True)
                    )
            else:
                listing, source_path = payload
                as_container = source_path.endswith("/")
                if no_glob:
                    dsrc = [
                        DataSource(
                            listing, listing.resolve_path(source_path), as_container
                        )
                    ]
                else:
                    dsrc = [
                        DataSource(listing, n, as_container)
                        for n in listing.expand_path(source_path)
                    ]
                node_groups.append(NodeGroup(listing, dsrc, source_path))

        return node_groups

    def unlist_source(self, uri: StorageURI) -> None:
        self.metastore.clone(uri=uri).mark_storage_not_indexed(uri)

    def storage_stats(self, uri: StorageURI) -> Optional[StorageStats]:
        """
        Returns tuple with storage stats: total number of rows and total dataset size.
        """
        partial_path = self.metastore.get_last_partial_path(uri)
        if partial_path is None:
            return None
        dataset = self.get_dataset(Storage.dataset_name(uri, partial_path))
        num_objects, size = self.warehouse.dataset_stats(
            dataset, dataset.latest_version
        )
        assert num_objects is not None
        assert size is not None

        return StorageStats(
            num_objects=num_objects,
            size=size,
        )

    def create_dataset(
        self,
        name: str,
        version: Optional[int] = None,
        query_script: str = "",
        create_rows: Optional[bool] = True,
        custom_columns: Sequence[Column] = (),
        validate_version: Optional[bool] = True,
        listing: Optional[bool] = False,
    ) -> "DatasetRecord":
        """
        Creates new dataset of a specific version.
        If dataset is not yet created, it will create it with version 1
        If version is None, then next unused version is created.
        If version is given, then it must be an unused version number.
        """
        if not listing and Client.is_data_source_uri(name):
            raise RuntimeError(
                "Cannot create dataset that starts with source prefix, e.g s3://"
            )
        default_version = 1
        try:
            dataset = self.get_dataset(name)
            default_version = dataset.next_version
        except DatasetNotFoundError:
            custom_column_types = {
                c.name: c.type.to_dict()
                for c in list(custom_columns)
                if isinstance(c.type, SQLType)
            }
            dataset = self.metastore.create_dataset(
                name,
                query_script=query_script,
                custom_column_types=custom_column_types,
                ignore_if_exists=True,
            )

        version = version or default_version

        if dataset.has_version(version):
            raise DatasetInvalidVersionError(
                f"Version {version} already exists in dataset {name}"
            )

        if validate_version and not dataset.is_valid_next_version(version):
            raise DatasetInvalidVersionError(
                f"Version {version} must be higher than the current latest one"
            )

        return self.create_new_dataset_version(
            dataset,
            version,
            query_script=query_script,
            create_rows_table=create_rows,
            custom_columns=custom_columns,
        )

    def create_new_dataset_version(
        self,
        dataset: DatasetRecord,
        version: int,
        sources="",
        query_script="",
        error_message="",
        error_stack="",
        script_output="",
        create_rows_table=True,
        custom_columns: Sequence[Column] = (),
    ) -> DatasetRecord:
        """
        Creates dataset version if it doesn't exist.
        If create_rows is False, dataset rows table will not be created
        """
        custom_column_types = {
            c.name: c.type.to_dict()
            for c in self.warehouse.schema.dataset_row_cls.dataset_default_columns()
            + list(custom_columns)
            if isinstance(c.type, SQLType)
        }

        dataset = self.metastore.create_dataset_version(
            dataset,
            version,
            status=DatasetStatus.PENDING,
            sources=sources,
            query_script=query_script,
            error_message=error_message,
            error_stack=error_stack,
            script_output=script_output,
            custom_column_types=custom_column_types,
            ignore_if_exists=True,
        )

        if create_rows_table:
            table_name = self.warehouse.dataset_table_name(dataset.name, version)
            self.warehouse.create_dataset_rows_table(
                table_name, custom_columns=custom_columns
            )

            self.update_dataset_version_with_warehouse_info(dataset, version)

        return dataset

    def update_dataset_version_with_warehouse_info(
        self, dataset: DatasetRecord, version: int, **kwargs
    ) -> None:
        dataset_version = dataset.get_version(version)

        values = {**kwargs}

        if not dataset_version.num_objects:
            num_objects, size = self.warehouse.dataset_stats(dataset, version)
            if num_objects != dataset_version.num_objects:
                values["num_objects"] = num_objects
            if size != dataset_version.size:
                values["size"] = size

        if not dataset_version.preview and (
            dataset_rows := list(
                self.ls_dataset_rows(
                    dataset.name,
                    version,
                    limit=20,
                    custom_columns=True,
                )
            )
        ):
            values["preview"] = [row.to_preview() for row in dataset_rows]

        if not values:
            return

        self.metastore.update_dataset_version(
            dataset,
            version,
            **values,
        )

    def update_dataset(
        self, dataset: DatasetRecord, conn=None, **kwargs
    ) -> DatasetRecord:
        """Updates dataset fields."""
        old_name = None
        new_name = None
        if "name" in kwargs and kwargs["name"] != dataset.name:
            old_name = dataset.name
            new_name = kwargs["name"]

        dataset = self.metastore.update_dataset(dataset, conn=conn, **kwargs)

        if old_name and new_name:
            # updating name must result in updating dataset table names as well
            for version in [v.version for v in dataset.versions]:
                self.warehouse.rename_dataset_table(
                    old_name,
                    new_name,
                    old_version=version,
                    new_version=version,
                )

        return dataset

    def remove_dataset_version(
        self, dataset: DatasetRecord, version: int, drop_rows: Optional[bool] = True
    ) -> None:
        """
        Deletes one single dataset version.
        If it was last version, it removes dataset completely
        """
        if not dataset.has_version(version):
            return
        dataset = self.metastore.remove_dataset_version(dataset, version)
        if drop_rows:
            self.warehouse.drop_dataset_rows_table(dataset, version)

    def get_temp_table_names(self) -> list[str]:
        return self.warehouse.get_temp_table_names()

    def cleanup_temp_tables(self, names: Iterable[str]) -> None:
        """
        Drop tables created temporarily when processing datasets.

        This should be implemented even if temporary tables are used to
        ensure that they are cleaned up as soon as they are no longer
        needed. When running the same `DatasetQuery` multiple times we
        may use the same temporary table names.
        """
        self.warehouse.cleanup_temp_tables(names)
        self.id_generator.delete_uris(names)

    def create_dataset_from_sources(
        self,
        name: str,
        sources: list[str],
        client_config=None,
        recursive=False,
    ) -> DatasetRecord:
        if not sources:
            raise ValueError("Sources needs to be non empty list")

        from dvcx.query import DatasetQuery

        dataset_queries = []
        for source in sources:
            if source.startswith(DATASET_PREFIX):
                dq = DatasetQuery(
                    name=source[len(DATASET_PREFIX) :],
                    catalog=self,
                    client_config=client_config,
                )
            else:
                dq = DatasetQuery(
                    path=source,
                    catalog=self,
                    client_config=client_config,
                    recursive=recursive,
                )

            dataset_queries.append(dq)

        # create union of all dataset queries created from sources
        dq = reduce(lambda ds1, ds2: ds1.union(ds2), dataset_queries)
        try:
            dq.save(name)
        except Exception as e:  # noqa: BLE001
            try:
                ds = self.get_dataset(name)
                self.metastore.update_dataset_status(
                    ds,
                    DatasetStatus.FAILED,
                    version=ds.latest_version,
                    error_message=DATASET_INTERNAL_ERROR_MESSAGE,
                    error_stack=traceback.format_exc(),
                )
                self.warehouse.drop_dataset_rows_table(ds, ds.latest_version)
                self.update_dataset_version_with_warehouse_info(
                    ds,
                    ds.latest_version,
                    sources="\n".join(sources),
                )
                raise
            except DatasetNotFoundError:
                raise e from None

        ds = self.get_dataset(name)

        self.update_dataset_version_with_warehouse_info(
            ds,
            ds.latest_version,
            sources="\n".join(sources),
        )

        return self.get_dataset(name)

    def register_new_dataset(
        self,
        source_dataset: DatasetRecord,
        source_version: int,
        target_name: str,
    ) -> DatasetRecord:
        target_dataset = self.metastore.create_dataset(
            target_name,
            query_script=source_dataset.query_script,
            custom_column_types=source_dataset.custom_column_types_serialized,
        )
        return self.register_dataset(source_dataset, source_version, target_dataset, 1)

    def register_dataset(
        self,
        dataset: DatasetRecord,
        version: int,
        target_dataset: DatasetRecord,
        target_version: Optional[int] = None,
    ) -> DatasetRecord:
        """
        Registers dataset version of one dataset as dataset version of another
        one (it can be new version of existing one).
        It also removes original dataset version
        """
        target_version = target_version or target_dataset.next_version

        if not target_dataset.is_valid_next_version(target_version):
            raise DatasetInvalidVersionError(
                f"Version {target_version} must be higher than the current latest one"
            )

        dataset_version = dataset.get_version(version)
        if not dataset_version:
            raise ValueError(f"Dataset {dataset.name} does not have version {version}")

        if not dataset_version.is_final_status():
            raise ValueError("Cannot register dataset version in non final status")

        # copy dataset version
        target_dataset = self.metastore.create_dataset_version(
            target_dataset,
            target_version,
            sources=dataset_version.sources,
            status=dataset_version.status,
            query_script=dataset_version.query_script,
            error_message=dataset_version.error_message,
            error_stack=dataset_version.error_stack,
            script_output=dataset_version.script_output,
            created_at=dataset_version.created_at,
            finished_at=dataset_version.finished_at,
            custom_column_types=dataset_version.custom_column_types_serialized,
            num_objects=dataset_version.num_objects,
            size=dataset_version.size,
            preview=dataset_version.preview,
        )
        # to avoid re-creating rows table, we are just renaming it for a new version
        # of target dataset
        self.warehouse.rename_dataset_table(
            dataset.name,
            target_dataset.name,
            old_version=version,
            new_version=target_version,
        )
        self.metastore.update_dataset_dependency_source(
            dataset,
            version,
            new_source_dataset=target_dataset,
            new_source_dataset_version=target_version,
        )

        if dataset.id == target_dataset.id:
            # we are updating the same dataset so we need to refresh it to have newly
            # added version in step before
            dataset = self.get_dataset(dataset.name)

        self.remove_dataset_version(dataset, version, drop_rows=False)

        return self.get_dataset(target_dataset.name)

    def get_dataset(self, name: str) -> DatasetRecord:
        return self.metastore.get_dataset(name)

    def get_remote_dataset(self, name: str, *, remote_config=None) -> DatasetRecord:
        remote_config = remote_config or get_remote_config(
            read_config(DVCXDir.find().root), remote=""
        )
        studio_client = StudioClient(
            remote_config["url"], remote_config["username"], remote_config["token"]
        )

        info_response = studio_client.dataset_info(name)
        if not info_response.ok:
            _raise_remote_error(info_response.message)

        dataset_info = info_response.data
        assert isinstance(dataset_info, dict)
        return DatasetRecord.from_dict(dataset_info)

    def get_dataset_dependencies(
        self, name: str, version: int, indirect=False
    ) -> list[Optional[DatasetDependency]]:
        dataset = self.get_dataset(name)

        direct_dependencies = self.metastore.get_direct_dataset_dependencies(
            dataset, version
        )

        if not indirect:
            return direct_dependencies

        for d in direct_dependencies:
            if not d:
                # dependency has been removed
                continue
            if d.is_dataset:
                # only datasets can have dependencies
                d.dependencies = self.get_dataset_dependencies(
                    d.name, int(d.version), indirect=indirect
                )

        return direct_dependencies

    def ls_datasets(self) -> Iterator[DatasetRecord]:
        datasets = self.metastore.list_datasets()
        for d in datasets:
            if not d.is_bucket_listing:
                yield d

    def ls_dataset_rows(
        self, name: str, version: int, offset=None, limit=None, custom_columns=False
    ) -> Iterator[DatasetRow]:
        dataset = self.get_dataset(name)

        yield from self.warehouse.get_dataset_rows(
            dataset,
            version,
            offset=offset,
            limit=limit,
            custom_columns=custom_columns,
        )

    def signed_url(self, source: str, path: str, client_config=None) -> str:
        client_config = client_config or self.client_config
        client, _ = self.parse_url(source, **client_config)
        return client.url(path)

    def export_dataset_table(
        self,
        bucket_uri: str,
        name: str,
        version: int,
        client_config=None,
    ) -> list[str]:
        dataset = self.get_dataset(name)

        return self.warehouse.export_dataset_table(
            bucket_uri, dataset, version, client_config
        )

    def dataset_table_export_file_names(self, name: str, version: int) -> list[str]:
        dataset = self.get_dataset(name)
        return self.warehouse.dataset_table_export_file_names(dataset, version)

    def dataset_stats(self, name: str, version: int) -> DatasetStats:
        """
        Returns tuple with dataset stats: total number of rows and total dataset size.
        """
        dataset = self.get_dataset(name)
        dataset_version = dataset.get_version(version)
        return DatasetStats(
            num_objects=dataset_version.num_objects,
            size=dataset_version.size,
        )

    def remove_dataset(
        self,
        name: str,
        version: Optional[int] = None,
        force: Optional[bool] = False,
    ):
        dataset = self.get_dataset(name)
        if not version and not force:
            raise ValueError(f"Missing dataset version from input for dataset {name}")
        if version and not dataset.has_version(version):
            raise DatasetInvalidVersionError(
                f"Dataset {name} doesn't have version {version}"
            )

        if version:
            self.remove_dataset_version(dataset, version)
            return

        while dataset.versions:
            version = dataset.versions[0].version
            self.remove_dataset_version(
                dataset,
                version,
            )

    def edit_dataset(
        self,
        name: str,
        new_name: Optional[str] = None,
        description: Optional[str] = None,
        labels: Optional[list[str]] = None,
    ) -> DatasetRecord:
        update_data = {}
        if new_name:
            update_data["name"] = new_name
        if description is not None:
            update_data["description"] = description
        if labels is not None:
            update_data["labels"] = labels  # type: ignore[assignment]

        dataset = self.get_dataset(name)
        return self.update_dataset(dataset, **update_data)

    def merge_datasets(
        self,
        src: DatasetRecord,
        dst: DatasetRecord,
        src_version: int,
        dst_version: Optional[int] = None,
    ) -> DatasetRecord:
        """
        Merges records from source to destination dataset.
        It will create new version
        of a dataset with records merged from old version and the source, unless
        existing version is specified for destination in which case it must
        be in non final status as datasets are immutable
        """
        if (
            dst_version
            and not dst.is_valid_next_version(dst_version)
            and dst.get_version(dst_version).is_final_status()
        ):
            raise DatasetInvalidVersionError(
                f"Version {dst_version} must be higher than the current latest one"
            )

        src_dep = self.get_dataset_dependencies(src.name, src_version)
        dst_dep = self.get_dataset_dependencies(
            dst.name,
            dst.latest_version,  # type: ignore[arg-type]
        )

        if dst.has_version(dst_version):  # type: ignore[arg-type]
            # case where we don't create new version, but append to the existing one
            self.warehouse.merge_dataset_rows(
                src,
                dst,
                src_version,
                dst_version=dst_version,  # type: ignore[arg-type]
            )
            merged_custom_column_types = {
                **src.custom_column_types_serialized,
                **dst.custom_column_types_serialized,
            }
            self.update_dataset(dst, custom_column_types=merged_custom_column_types)
            self.update_dataset_version_with_warehouse_info(
                dst,
                dst_version,  # type: ignore[arg-type]
                custom_column_types=merged_custom_column_types,
            )
            for dep in src_dep:
                if dep and dep not in dst_dep:
                    self.metastore.add_dependency(
                        dep,
                        dst.name,
                        dst_version,  # type: ignore[arg-type]
                    )
        else:
            # case where we create new version of merged results
            src_dr = self.warehouse.dataset_rows(src, src_version)
            dst_dr = self.warehouse.dataset_rows(dst)

            merge_result_columns = list(
                {
                    c.name: c for c in list(src_dr.table.c) + list(dst_dr.table.c)
                }.values()
            )
            custom_columns = [
                c
                for c in merge_result_columns
                if c.name not in DATASET_CORE_COLUMN_NAMES
            ]

            dst_version = dst_version or dst.next_version
            dst = self.create_new_dataset_version(
                dst, dst_version, custom_columns=custom_columns
            )
            self.warehouse.merge_dataset_rows(
                src,
                dst,
                src_version,
                dst_version,
            )
            self.update_dataset_version_with_warehouse_info(dst, dst_version)
            for dep in set(src_dep + dst_dep):
                if dep:
                    self.metastore.add_dependency(dep, dst.name, dst_version)

        return dst

    def open_object(self, row: DatasetRow, use_cache: bool = True, **config: Any):
        config = config or self.client_config
        client = self.get_client(row.source, **config)
        return client.open_object(row.as_uid(), use_cache=use_cache)

    def ls(
        self,
        sources: list[str],
        fields: Iterable[str],
        ttl=TTL_INT,
        update=False,
        skip_indexing=False,
        *,
        client_config=None,
    ) -> Iterator[tuple[DataSource, Iterable[tuple]]]:
        data_sources = self.enlist_sources(
            sources,
            ttl,
            update,
            skip_indexing=skip_indexing,
            client_config=client_config or self.client_config,
        )

        for source in data_sources:  # type: ignore [union-attr]
            yield source, source.ls(fields)

    def ls_storage_uris(self) -> Iterator[str]:
        yield from self.metastore.get_all_storage_uris()

    def get_storage(self, uri: StorageURI) -> Storage:
        return self.metastore.get_storage(uri)

    def ls_storages(self) -> list[Storage]:
        return self.metastore.list_storages()

    def pull_dataset(
        self,
        dataset_uri: str,
        output: Optional[str] = None,
        no_cp: bool = False,
        force: bool = False,
        edvcx: bool = False,
        edvcx_file: Optional[str] = None,
        *,
        client_config=None,
        remote_config=None,
    ) -> None:
        # TODO add progress bar https://github.com/iterative/dvcx/issues/750
        # TODO copy correct remote dates https://github.com/iterative/dvcx/issues/new
        # TODO compare dataset stats on remote vs local pull to assert it's ok
        def _instantiate_dataset():
            if no_cp:
                return
            self.cp(
                [dataset_uri],
                output,
                force=force,
                no_edvcx_file=not edvcx,
                edvcx_file=edvcx_file,
                client_config=client_config,
            )
            print(f"Dataset {dataset_uri} instantiated locally to {output}")

        if not output and not no_cp:
            raise ValueError("Please provide output directory for instantiation")

        client_config = client_config or self.client_config
        remote_config = remote_config or get_remote_config(
            read_config(DVCXDir.find().root), remote=""
        )

        studio_client = StudioClient(
            remote_config["url"], remote_config["username"], remote_config["token"]
        )

        try:
            remote_dataset_name, version = parse_dataset_uri(dataset_uri)
        except Exception as e:
            raise DVCXError("Error when parsing dataset uri") from e

        dataset = None
        try:
            dataset = self.get_dataset(remote_dataset_name)
        except DatasetNotFoundError:
            # we will create new one if it doesn't exist
            pass

        remote_dataset = self.get_remote_dataset(
            remote_dataset_name, remote_config=remote_config
        )
        # if version is not specified in uri, take the latest one
        if not version:
            version = remote_dataset.latest_version
            print(f"Version not specified, pulling the latest one (v{version})")
            # updating dataset uri with latest version
            dataset_uri = create_dataset_uri(remote_dataset_name, version)

        assert version

        if dataset and dataset.has_version(version):
            print(f"Local copy of dataset {dataset_uri} already present")
            _instantiate_dataset()
            return

        try:
            remote_dataset_version = remote_dataset.get_version(version)
        except (ValueError, StopIteration) as exc:
            raise DVCXError(
                f"Dataset {remote_dataset_name} doesn't have version {version}"
                " on server"
            ) from exc

        stats_response = studio_client.dataset_stats(remote_dataset_name, version)
        if not stats_response.ok:
            _raise_remote_error(stats_response.message)
        dataset_stats = stats_response.data

        dataset_save_progress_bar = tqdm(
            desc=f"Saving dataset {dataset_uri} locally: ",
            unit=" rows",
            unit_scale=True,
            unit_divisor=1000,
            total=dataset_stats.num_objects,  # type: ignore [union-attr]
        )

        custom_column_types = DatasetRecord.parse_custom_column_types(
            remote_dataset_version.custom_column_types,
        )
        custom_columns = [
            sa.Column(c_name, c_type)
            for c_name, c_type in custom_column_types.items()
            if c_name not in DATASET_CORE_COLUMN_NAMES
        ]

        # creating new dataset (version) locally
        dataset = self.create_dataset(
            remote_dataset_name,
            version,
            query_script=remote_dataset_version.query_script,
            create_rows=True,
            custom_columns=custom_columns,
            validate_version=False,
        )

        # asking remote to export dataset rows table to s3 and to return signed
        # urls of exported parts, which are in parquet format
        export_response = studio_client.export_dataset_table(
            remote_dataset_name, version
        )
        if not export_response.ok:
            _raise_remote_error(export_response.message)

        signed_urls = export_response.data

        if signed_urls:
            shuffle(signed_urls)

            rows_fetcher = DatasetRowsFetcher(
                self.metastore.clone(),
                self.warehouse.clone(),
                remote_config,
                dataset.name,
                version,
                custom_column_types,
            )
            try:
                rows_fetcher.run(
                    batched(
                        signed_urls,
                        math.ceil(len(signed_urls) / PULL_DATASET_MAX_THREADS),
                    ),
                    dataset_save_progress_bar,
                )
            except:
                self.remove_dataset(dataset.name, version)
                raise

        dataset = self.metastore.update_dataset_status(
            dataset,
            DatasetStatus.COMPLETE,
            version=version,
            error_message=remote_dataset.error_message,
            error_stack=remote_dataset.error_stack,
            script_output=remote_dataset.error_stack,
        )
        self.update_dataset_version_with_warehouse_info(dataset, version)

        dataset_save_progress_bar.close()
        print(f"Dataset {dataset_uri} saved locally")

        _instantiate_dataset()

    def clone(
        self,
        sources: list[str],
        output: str,
        force: bool = False,
        update: bool = False,
        recursive: bool = False,
        no_glob: bool = False,
        no_cp: bool = False,
        edvcx: bool = False,
        edvcx_file: Optional[str] = None,
        ttl: int = TTL_INT,
        *,
        client_config=None,
    ) -> None:
        """
        This command takes cloud path(s) and duplicates files and folders in
        them into the dataset folder.
        It also adds those files to a dataset in database, which is
        created if doesn't exist yet
        Optionally, it creates a .edvcx file
        """
        if not no_cp or edvcx:
            self.cp(
                sources,
                output,
                force=force,
                update=update,
                recursive=recursive,
                no_glob=no_glob,
                edvcx_only=no_cp,
                no_edvcx_file=not edvcx,
                edvcx_file=edvcx_file,
                ttl=ttl,
                client_config=client_config,
            )
        else:
            # since we don't call cp command, which does listing implicitly,
            # it needs to be done here
            self.enlist_sources(
                sources,
                ttl,
                update,
                client_config=client_config or self.client_config,
            )

        self.create_dataset_from_sources(
            output, sources, client_config=client_config, recursive=recursive
        )

    def apply_udf(
        self,
        udf_location: str,
        source: str,
        target_name: str,
        parallel: Optional[int] = None,
        params: Optional[str] = None,
    ):
        from dvcx.query import DatasetQuery

        if source.startswith(DATASET_PREFIX):
            ds = DatasetQuery(name=source[len(DATASET_PREFIX) :], catalog=self)
        else:
            ds = DatasetQuery(path=source, catalog=self)
        udf = import_object(udf_location)
        if params:
            args, kwargs = parse_params_string(params)
            udf = udf(*args, **kwargs)
        ds.add_signals(udf, parallel=parallel).save(target_name)

    def query(
        self,
        query_script: str,
        envs: Optional[Mapping[str, str]] = None,
        python_executable: Optional[str] = None,
        save: bool = False,
        save_as: Optional[str] = None,
        preview_limit: Optional[int] = 10,
        preview_offset: int = 0,
        preview_columns: Optional[list[str]] = None,
        capture_output: bool = True,
        output_hook: Callable[[str], None] = noop,
        params: Optional[dict[str, str]] = None,
    ) -> QueryResult:
        """
        Method to run custom user Python script to run a query and, as result,
        creates new dataset from the results of a query.
        Returns tuple of result dataset and script output.

        Constraints on query script:
            1. dvcx.query.DatasetQuery should be used in order to create query
            for a dataset
            2. There should not be any .save() call on DatasetQuery since the idea
            is to create only one dataset as the outcome of the script
            3. Last statement must be an instance of DatasetQuery

        If save is set to True, we are creating new dataset with results
        from dataset query. If it's set to False, we will just print results
        without saving anything

        Example of query script:
            from dvcx.query import DatasetQuery, C
            DatasetQuery('s3://ldb-public/remote/datasets/mnist-tiny/').filter(
                C.size > 1000
            )
        """
        from dvcx.query.dataset import ExecutionResult

        try:
            query_script_compiled = self.compile_query_script(query_script)
        except Exception as exc:
            raise QueryScriptCompileError(
                f"Query script failed to compile, reason: {exc}"
            ) from exc

        if save_as and save_as.startswith(QUERY_DATASET_PREFIX):
            raise ValueError(
                f"Cannot use {QUERY_DATASET_PREFIX} prefix for dataset name"
            )

        r, w = os.pipe()
        if os.name == "nt":
            import msvcrt

            os.set_inheritable(w, True)

            startupinfo = subprocess.STARTUPINFO()  # type: ignore[attr-defined]
            handle = msvcrt.get_osfhandle(w)  # type: ignore[attr-defined]
            startupinfo.lpAttributeList["handle_list"].append(handle)
            kwargs: dict[str, Any] = {"startupinfo": startupinfo}
        else:
            handle = w
            kwargs = {"pass_fds": [w]}

        envs = dict(envs or os.environ)
        envs.update(
            {
                "DVCX_QUERY_PARAMS": json.dumps(params or {}),
                "DVCX_QUERY_PREVIEW_ARGS": json.dumps(
                    {
                        "limit": preview_limit,
                        "offset": preview_offset,
                        "columns": preview_columns,
                    }
                ),
                "DVCX_QUERY_SAVE": "1" if save else "",
                "DVCX_QUERY_SAVE_AS": save_as or "",
                "PYTHONUNBUFFERED": "1",
                "DVCX_OUTPUT_FD": str(handle),
            },
        )

        python_executable = python_executable or sys.executable
        with subprocess.Popen(
            [python_executable, "-c", query_script_compiled],  # noqa: S603
            env=envs,
            stdout=subprocess.PIPE if capture_output else None,
            stderr=subprocess.STDOUT if capture_output else None,
            bufsize=1,
            text=True,
            **kwargs,
        ) as proc:
            os.close(w)

            out = proc.stdout
            _lines: list[str] = []
            ctx = print_and_capture(out, output_hook) if out else nullcontext(_lines)

            with ctx as lines, open(r) as f:
                response_text = ""
                while proc.poll() is None:
                    response_text += f.readline()
                    time.sleep(0.1)
                response_text += f.readline()

        output = "".join(lines)

        if proc.returncode:
            if proc.returncode == QUERY_SCRIPT_CANCELED_EXIT_CODE:
                raise QueryScriptCancelError(
                    "Query script was canceled by user",
                    return_code=proc.returncode,
                    output=output,
                )
            if proc.returncode == QUERY_SCRIPT_INVALID_LAST_STATEMENT_EXIT_CODE:
                raise QueryScriptRunError(
                    "Last line in a script was not an instance of DatasetQuery",
                    return_code=proc.returncode,
                    output=output,
                )
            raise QueryScriptRunError(
                f"Query script exited with error code {proc.returncode}",
                return_code=proc.returncode,
                output=output,
            )

        try:
            response = json.loads(response_text)
        except ValueError:
            response = {}
        exec_result = ExecutionResult(**response)

        dataset: Optional[DatasetRecord] = None
        version: Optional[int] = None
        if save or save_as:
            if not exec_result.dataset:
                raise QueryScriptDatasetNotFound(
                    "No dataset found after running Query script",
                    output=output,
                )
            name, version = exec_result.dataset
            # finding returning dataset
            try:
                dataset = self.get_dataset(name)
                dataset.get_version(version)
            except (DatasetNotFoundError, ValueError) as e:
                raise QueryScriptDatasetNotFound(
                    "No dataset found after running Query script",
                    output=output,
                ) from e

            dataset = self.update_dataset(
                dataset,
                script_output=output,
                query_script=query_script,
            )
            self.update_dataset_version_with_warehouse_info(
                dataset,
                version,
                script_output=output,
                query_script=query_script,
            )

        return QueryResult(
            dataset=dataset,
            version=version,
            output=output,
            preview=exec_result.preview,
        )

    def cp(
        self,
        sources: list[str],
        output: str,
        force: bool = False,
        update: bool = False,
        recursive: bool = False,
        edvcx_file: Optional[str] = None,
        edvcx_only: bool = False,
        no_edvcx_file: bool = False,
        no_glob: bool = False,
        ttl: int = TTL_INT,
        *,
        client_config=None,
    ) -> list[dict[str, Any]]:
        """
        This function copies files from cloud sources to local destination directory
        If cloud source is not indexed, or has expired index, it runs indexing
        It also creates .edvcx file by default, if not specified differently
        """
        client_config = client_config or self.client_config
        node_groups = self.enlist_sources_grouped(
            sources,
            ttl,
            update,
            no_glob,
            client_config=client_config,
        )

        always_copy_dir_contents, copy_to_filename = prepare_output_for_cp(
            node_groups, output, force, edvcx_only, no_edvcx_file
        )
        dataset_file = check_output_dataset_file(
            output, force, edvcx_file, no_edvcx_file
        )

        total_size, total_files = collect_nodes_for_cp(node_groups, recursive)

        if total_files == 0:
            # Nothing selected to cp
            return []

        desc_max_len = max(len(output) + 16, 19)
        bar_format = (
            "{desc:<"
            f"{desc_max_len}"
            "}{percentage:3.0f}%|{bar}| {n_fmt:>5}/{total_fmt:<5} "
            "[{elapsed}<{remaining}, {rate_fmt:>8}]"
        )

        if not edvcx_only:
            with get_download_bar(bar_format, total_size) as pbar:
                for node_group in node_groups:
                    node_group.download(recursive=recursive, pbar=pbar)

        instantiate_node_groups(
            node_groups,
            output,
            bar_format,
            total_files,
            force,
            recursive,
            edvcx_only,
            always_copy_dir_contents,
            copy_to_filename,
        )
        if no_edvcx_file:
            return []

        metafile_data = compute_metafile_data(node_groups)
        if metafile_data:
            # Don't write the metafile if nothing was copied
            print(f"Creating '{dataset_file}'")
            with open(dataset_file, "w", encoding="utf-8") as fd:
                yaml.dump(metafile_data, fd, sort_keys=False)

        return metafile_data

    def du(
        self,
        sources,
        depth=0,
        ttl=TTL_INT,
        update=False,
        *,
        client_config=None,
    ) -> Iterable[tuple[str, float]]:
        sources = self.enlist_sources(
            sources,
            ttl,
            update,
            client_config=client_config or self.client_config,
        )

        def du_dirs(src, node, subdepth):
            if subdepth > 0:
                subdirs = src.listing.get_nodes_by_parent_path(node.path, type="dir")
                for sd in subdirs:
                    yield from du_dirs(src, sd, subdepth - 1)
            yield (
                src.get_node_full_path(node),
                src.listing.du(node)[0],
            )

        for src in sources:
            yield from du_dirs(src, src.node, depth)

    def find(
        self,
        sources,
        ttl=TTL_INT,
        update=False,
        names=None,
        inames=None,
        paths=None,
        ipaths=None,
        size=None,
        typ=None,
        columns=None,
        *,
        client_config=None,
    ) -> Iterator[str]:
        sources = self.enlist_sources(
            sources,
            ttl,
            update,
            client_config=client_config or self.client_config,
        )
        if not columns:
            columns = ["path"]
        field_set = set()
        for column in columns:
            if column == "du":
                field_set.add("dir_type")
                field_set.add("size")
                field_set.add("parent")
                field_set.add("name")
            elif column == "name":
                field_set.add("name")
            elif column == "owner":
                field_set.add("owner_name")
            elif column == "path":
                field_set.add("dir_type")
                field_set.add("parent")
                field_set.add("name")
            elif column == "size":
                field_set.add("size")
            elif column == "type":
                field_set.add("dir_type")
        fields = list(field_set)
        field_lookup = {f: i for i, f in enumerate(fields)}
        for src in sources:
            results = src.listing.find(
                src.node, fields, names, inames, paths, ipaths, size, typ
            )
            for row in results:
                yield "\t".join(
                    find_column_to_str(row, field_lookup, src, column)
                    for column in columns
                )

    def index(
        self,
        sources,
        ttl=TTL_INT,
        update=False,
        *,
        client_config=None,
    ) -> None:
        root_sources = [
            src for src in sources if Client.get_implementation(src).is_root_url(src)
        ]
        non_root_sources = [
            src
            for src in sources
            if not Client.get_implementation(src).is_root_url(src)
        ]

        client_config = client_config or self.client_config

        # for root sources (e.g s3://) we are just getting all buckets and
        # saving them as storages, without further indexing in each bucket
        for source in root_sources:
            for bucket in Client.get_implementation(source).ls_buckets(**client_config):
                client = self.get_client(bucket.uri, **client_config)
                print(f"Registering storage {client.uri}")
                self.metastore.create_storage_if_not_registered(client.uri)

        self.enlist_sources(
            non_root_sources,
            ttl,
            update,
            client_config=client_config,
            only_index=True,
        )

    def find_stale_storages(self) -> None:
        self.metastore.find_stale_storages()



"""
This script provides an example of the dir expansion query available
for datasets and storage indices. It prints the results, some summary
stats and verifies that the result is the same both for a dataset and
for a storage index with the same entries.
"""

import sys

import sqlalchemy as sa

from dvcx.catalog import get_catalog
from dvcx.error import DatasetNotFoundError
from dvcx.query import DatasetQuery

show_ids = False
if len(sys.argv) > 1:
    if sys.argv[1] == "--show-ids":
        show_ids = True
    else:
        raise RuntimeError(f"Invalid arg: {sys.argv[1]}")

source = "gcs://dvcx-datasets"
catalog = get_catalog()
metastore = catalog.metastore
warehouse = catalog.warehouse

try:
    dataset = metastore.get_dataset("ds1")
except DatasetNotFoundError:
    DatasetQuery(source).save("ds1")
    dataset = metastore.get_dataset("ds1")

partial_id = metastore.get_valid_partial_id(source, "")
n = warehouse.nodes_table(source, partial_id)
dr = warehouse.dataset_rows(dataset, version=dataset.latest_version)


def get_values_to_compare(query):
    q = query.dir_expansion().subquery()
    return list(
        warehouse.db.execute(
            sa.select(
                q.c.source,
                q.c.parent,
                q.c.name,
                q.c.version,
                q.c.vtype,
                q.c.is_dir,
                q.c.location,
            )
        )
    )


dq = dr.dir_expansion()
result = warehouse.db.execute(dq)
for id_, vtype, is_dir, source, parent, name, version, _loc in result:
    id_str = f"{id_!r:12} " if show_ids else ""
    print(f"{id_str}{vtype!r:6} {version!r} {is_dir} {source!r} {parent!r} {name!r}")

print()
print("num dir expansion rows: ", len(list(warehouse.db.execute(dq))))
print(
    "num dir expansion files:",
    len(list(warehouse.db.execute(dq.having(~dq.selected_columns.is_dir)))),
)
print("num table rows:         ", len(list(warehouse.db.execute(dr.select()))))
print("same values:            ", get_values_to_compare(n) == get_values_to_compare(dr))



import io

from dvcx.lib.cached_stream import ProgressiveCacheStream
from dvcx.lib.file import File

DATA = b"qwertyuiop"
FILE_UID = File(name="name").get_uid()


def test_close(catalog):
    input = io.BytesIO()
    input.write(DATA)
    input.seek(0)

    with ProgressiveCacheStream(input, len(DATA), catalog, FILE_UID) as stream:
        stream.close()
        with open(catalog.cache.get_path(FILE_UID), "rb") as f:
            assert f.read() == DATA


def test_read_by_parts(catalog):
    input = io.BytesIO()
    input.write(DATA)
    input.seek(0)

    with ProgressiveCacheStream(input, len(DATA), catalog, FILE_UID) as stream:
        assert stream.read(1)[0] == DATA[0]
        assert stream.read(5) == DATA[1:6]
        assert stream.read(1)[0] == DATA[6]
        assert stream.read(1024) == DATA[7:]

        stream.close()
        with open(catalog.cache.get_path(FILE_UID), "rb") as f:
            assert f.read() == DATA


def test_seek_forward(catalog):
    input = io.BytesIO()
    input.write(DATA)
    input.seek(0)

    with ProgressiveCacheStream(input, len(DATA), catalog, FILE_UID) as stream:
        stream.seek(4)
        assert stream.read(1)[0] == DATA[4]

        stream.close()
        with open(catalog.cache.get_path(FILE_UID), "rb") as f:
            assert f.read() == DATA


def test_seek_backward(catalog):
    input = io.BytesIO()
    input.write(DATA)
    input.seek(0)

    with ProgressiveCacheStream(input, len(DATA), catalog, FILE_UID) as stream:
        stream.read(8)

        stream.seek(3)
        assert stream.tell() == 3
        assert stream.read(3) == DATA[3:6]

        stream.close()
        with open(catalog.cache.get_path(FILE_UID), "rb") as f:
            assert f.read() == DATA


def test_seek_backward_with_intersection(catalog):
    input = io.BytesIO()
    input.write(DATA)
    input.seek(0)

    with ProgressiveCacheStream(input, len(DATA), catalog, FILE_UID) as stream:
        stream.read(8)

        stream.seek(3)
        assert stream.tell() == 3
        assert stream.read(6) == DATA[3:9]

        stream.close()
        with open(catalog.cache.get_path(FILE_UID), "rb") as f:
            assert f.read() == DATA



from typing import Any

from adlfs import AzureBlobFileSystem

from dvcx.node import Entry

from .fsspec import DELIMITER, Client


class AzureClient(Client):
    FS_CLASS = AzureBlobFileSystem
    PREFIX = "az://"
    protocol = "az"

    def convert_info(self, v: dict[str, Any], parent: str) -> Entry:
        version_id = v.get("version_id")
        name = v.get("name", "").split(DELIMITER)[-1]
        if version_id:
            version_suffix = f"?versionid={version_id}"
            if name.endswith(version_suffix):
                name = name[: -len(version_suffix)]
        return Entry.from_file(
            parent=parent,
            name=name,
            etag=v.get("etag", "").strip('"'),
            version=version_id or "",
            is_latest=version_id is None or bool(v.get("is_current_version")),
            last_modified=v["last_modified"],
            size=v.get("size", ""),
        )



from typing import Optional

import numpy as np
from pydantic import Field

from dvcx.lib.feature import Feature
from dvcx.lib.feature_udf import FeatureAggregator
from dvcx.lib.file import File, FileInfo
from dvcx.lib.utils import DvcxError
from dvcx.query import C
from dvcx.sql.functions import path

try:
    import pandas as pd
except ImportError:
    pd = None


class WDSMetaError(DvcxError):
    def __init__(self, msg):
        super().__init__(msg)


def parse_wds_meta(ds):
    return ds.aggregate(MergeParquetAndNpz(), partition_by=path.file_stem(C.name))


class LaionMeta(Feature):
    uid: str = Field(default="")
    url: str = Field(default="")
    text: str = Field(default="")
    original_width: int = Field(default=-1)
    original_height: int = Field(default=-1)
    clip_b32_similarity_score: float = Field(default=0.0)
    clip_l14_similarity_score: float = Field(default=0.0)
    face_bboxes: Optional[list[list[float]]] = Field(default=None)
    sha256: str = Field(default="")

    b32_img: list[float] = Field(default=None)
    b32_txt: list[float] = Field(default=None)
    l14_img: list[float] = Field(default=None)
    l14_txt: list[float] = Field(default=None)
    dedup: list[float] = Field(default=None)


class MergeParquetAndNpz(FeatureAggregator):
    def __init__(self):
        super().__init__(File, [FileInfo, LaionMeta])

    def process(self, args):
        stream_npz, stream_pq = self.get_meta_streams(args)

        with stream_pq.open() as fd_pq, stream_npz.open() as fd_npz:
            df = pd.read_parquet(fd_pq)
            npz_file = np.load(fd_npz)

            yield from self.parse_metafiles(npz_file, df, stream_pq)

    @staticmethod
    def get_meta_streams(args):
        if len(args) != 2:
            filenames = " ".join([f.get_full_path() for f in args])
            raise WDSMetaError(f"npz-parquet pair mismatch: {filenames}")
        stream_pq = args[0]
        stream_npz = args[1]
        if args[0].get_file_ext() != "parquet":
            stream_pq, stream_npz = stream_npz, stream_pq
        return stream_npz, stream_pq

    @staticmethod
    def parse_metafiles(npz_file, df, stream_pq):
        b32_img = npz_file["b32_img"]
        b32_txt = npz_file["b32_txt"]
        l14_img = npz_file["l14_img"]
        l14_txt = npz_file["l14_txt"]
        dedup = npz_file["dedup"]
        for idx, (_, row) in enumerate(df.iterrows()):
            if idx > 1000:
                break

            fstream = FileInfo(
                name=str(idx),
                source=stream_pq.source,
                parent=f"{stream_pq.get_full_name()}",
                version=stream_pq.version,
                etag=f"{stream_pq.etag}_{stream_pq.etag}",
                vtype="parquet",
            )

            pq_dict = row.to_dict()
            npz_dict = {
                "b32_img": b32_img[idx],
                "b32_txt": b32_txt[idx],
                "l14_img": l14_img[idx],
                "l14_txt": l14_txt[idx],
                "dedup": dedup[idx],
            }

            yield fstream, LaionMeta(**(pq_dict | npz_dict))



import pytest
from sqlalchemy import Integer

from dvcx.dataset import DatasetRow
from dvcx.query import udf
from dvcx.query.batch import DatasetRowsBatch
from dvcx.query.schema import ColumnParameter


def test_udf_single_signal():
    @udf(("id", "size"), {"mul": Integer})
    def t(a, b):
        return (a * b,)

    row = DatasetRow(
        id=6,
        vtype="",
        dir_type=1,
        parent="",
        name="obj",
        last_modified=None,
        etag="",
        version="",
        is_latest=True,
        size=7,
        owner_name="",
        owner_id="",
        source="",
        random=1234,
        location=None,
    )
    result = t(None, row)
    assert result[0]["mul"] == (42)


def test_udf_multiple_signals():
    @udf(("id", "size"), {"mul": Integer, "sum": Integer})
    def t(a, b):
        return (a * b, a + b)

    row = DatasetRow(
        id=6,
        vtype="",
        dir_type=1,
        parent="",
        name="obj",
        last_modified=None,
        etag="",
        version="",
        is_latest=True,
        size=7,
        owner_name="",
        owner_id="",
        source="",
        random=1234,
        location=None,
    )
    result = t(None, row)
    assert result[0] == {"id": 6, "mul": 42, "sum": 13}


def test_udf_batching():
    @udf(("id", "size"), {"mul": Integer}, batch=4)
    def t(vals):
        return [(a * b,) for (a, b) in vals]

    inputs = list(zip(range(1, 11), range(21, 31)))
    results = []
    for size, row_id in inputs:
        row = DatasetRow(
            id=row_id,
            vtype="",
            dir_type=1,
            parent="",
            name="obj",
            last_modified=None,
            etag="",
            version="",
            is_latest=True,
            size=size,
            owner_name="",
            owner_id="",
            source="",
            random=1234,
            location=None,
        )
        batch = DatasetRowsBatch([row])
        result = t(None, batch)
        if result:
            assert len(result) == 1  # Matches batch size.
            results.extend(result)

    assert len(results) == len(inputs)
    assert results == [{"id": b, "mul": a * b} for (a, b) in inputs]


def test_stateful_udf():
    @udf(("size",), {"sum": Integer}, method="sum")
    class MyUDF:
        def __init__(self, constant):
            self.constant = constant

        def sum(self, size):
            return (self.constant + size,)

    udf_inst = MyUDF(5)()
    inputs = range(1, 11)
    results = []
    for size in inputs:
        row = DatasetRow(
            id=5,
            vtype="",
            dir_type=1,
            parent="",
            name="obj",
            last_modified=None,
            etag="",
            version="",
            is_latest=True,
            size=size,
            owner_name="",
            owner_id="",
            source="",
            random=1234,
            location=None,
        )
        results.extend(udf_inst(None, row))

    assert len(results) == len(inputs)
    assert results == [{"id": 5, "sum": 5 + size} for size in inputs]


@pytest.mark.parametrize("param", ["foo", ("foo",)])
def test_udf_api(param):
    func = lambda x: x  # noqa: E731
    result = udf(param, {"bar": Integer}, batch=42)(func)
    assert result.func is func
    assert result.properties.params == [ColumnParameter("foo")]
    assert result.properties.output == {"bar": Integer}
    assert result.properties.batch == 42


def test_udf_error():
    with pytest.raises(TypeError):

        @udf(params=("name",), output=("name_len",))
        def name_len(name):
            return len(name)



import logging
import sqlite3
from collections.abc import Iterable
from datetime import MAXYEAR, MINYEAR, datetime, timezone
from types import MappingProxyType
from typing import Callable, Optional

import sqlalchemy as sa
import ujson
from sqlalchemy.dialects import sqlite
from sqlalchemy.ext.compiler import compiles
from sqlalchemy.sql.elements import literal
from sqlalchemy.sql.expression import case
from sqlalchemy.sql.functions import func

from dvcx.sql.functions import array, conditional, random, string
from dvcx.sql.functions import path as sql_path
from dvcx.sql.selectable import Values, base_values_compiler
from dvcx.sql.sqlite.types import (
    SQLiteTypeConverter,
    SQLiteTypeReadConverter,
    register_type_converters,
)
from dvcx.sql.types import (
    TypeDefaults,
    register_backend_types,
    register_type_defaults,
    register_type_read_converters,
)

logger = logging.getLogger("dvcx")

_registered_function_creators: dict[str, Callable[[sqlite3.Connection], None]] = {}
registered_function_creators = MappingProxyType(_registered_function_creators)

_compiler_hooks: dict[str, Callable] = {}

sqlite_dialect = sqlite.dialect(paramstyle="named")

setup_is_complete: bool = False

slash = literal("/")
empty_str = literal("")
dot = literal(".")


def setup():
    global setup_is_complete  # noqa: PLW0603
    if setup_is_complete:
        return

    # sqlite 3.31.1 is the earliest version tested in CI
    if sqlite3.sqlite_version_info < (3, 31, 1):
        logger.warning(
            "Possible sqlite incompatibility. The earliest tested version of "
            "sqlite is 3.31.1 but you have %s",
            sqlite3.sqlite_version,
        )

    # We want to show tracebacks for user-defined functions
    sqlite3.enable_callback_tracebacks(True)
    sqlite3.register_adapter(datetime, adapt_datetime)
    sqlite3.register_converter("datetime", convert_datetime)

    register_type_converters()
    register_backend_types("sqlite", SQLiteTypeConverter())
    register_type_read_converters("sqlite", SQLiteTypeReadConverter())
    register_type_defaults("sqlite", TypeDefaults())

    compiles(sql_path.parent, "sqlite")(compile_path_parent)
    compiles(sql_path.name, "sqlite")(compile_path_name)
    compiles(sql_path.file_stem, "sqlite")(compile_path_file_stem)
    compiles(sql_path.file_ext, "sqlite")(compile_path_file_ext)
    compiles(array.cosine_distance, "sqlite")(compile_cosine_distance)
    compiles(array.euclidean_distance, "sqlite")(compile_euclidean_distance)
    compiles(array.length, "sqlite")(compile_array_length)
    compiles(string.length, "sqlite")(compile_string_length)
    compiles(string.split, "sqlite")(compile_string_split)
    compiles(conditional.greatest, "sqlite")(compile_greatest)
    compiles(conditional.least, "sqlite")(compile_least)
    compiles(Values, "sqlite")(compile_values)
    compiles(random.rand, "sqlite")(compile_rand)

    register_user_defined_sql_functions()
    setup_is_complete = True


def run_compiler_hook(name):
    try:
        hook = _compiler_hooks[name]
    except KeyError:
        return
    hook()


def functions_exist(
    names: Iterable[str], connection: Optional[sqlite3.Connection] = None
) -> bool:
    """
    Returns True if all function names are defined for the given connection.
    """

    names = list(names)
    for n in names:
        if not isinstance(n, str):
            raise TypeError(
                "functions_exist(): names argument must contain str values. "
                f"Found value of type {type(n).__name__}: {n!r}"
            )

    if connection is None:
        connection = sqlite3.connect(":memory:")

    if not names:
        return True
    column1 = sa.column("column1", sa.String)
    func_name_query = column1.not_in(
        sa.select(sa.column("name", sa.String)).select_from(func.pragma_function_list())
    )
    query = (
        sa.select(func.count() == 0)
        .select_from(sa.values(column1).data([(n,) for n in names]))
        .where(func_name_query)
    )
    comp = query.compile(dialect=sqlite_dialect)
    args = (comp.string, comp.params) if comp.params else (comp.string,)
    return bool(connection.execute(*args).fetchone()[0])


def create_user_defined_sql_functions(connection):
    for function_creator in registered_function_creators.values():
        function_creator(connection)


def missing_vector_function(name, exc):
    def unavailable_func(*args):
        raise ImportError(
            f"Missing dependencies for SQL vector function, {name}\n"
            "To install run:\n\n"
            "  pip install 'dvcx[vector]'\n"
        ) from exc

    return unavailable_func


def sqlite_string_split(string: str, sep: str, maxsplit: int = -1) -> str:
    return ujson.dumps(string.split(sep, maxsplit))


def register_user_defined_sql_functions() -> None:
    # Register optional functions if we have the necessary dependencies
    # and otherwise register functions that will raise an exception with
    # installation instructions
    try:
        from .vector import cosine_distance, euclidean_distance
    except ImportError as exc:
        # We want to throw an exception when trying to compile these
        # functions and also if the functions are called using raw SQL.
        cosine_distance = missing_vector_function("cosine_distance", exc)
        euclidean_distance = missing_vector_function("euclidean_distance", exc)
        _compiler_hooks["cosine_distance"] = cosine_distance
        _compiler_hooks["euclidean_distance"] = euclidean_distance

    def create_vector_functions(conn):
        conn.create_function("cosine_distance", 2, cosine_distance, deterministic=True)
        conn.create_function(
            "euclidean_distance", 2, euclidean_distance, deterministic=True
        )

    _registered_function_creators["vector_functions"] = create_vector_functions

    def create_string_functions(conn):
        conn.create_function("split", 2, sqlite_string_split, deterministic=True)
        conn.create_function("split", 3, sqlite_string_split, deterministic=True)

    _registered_function_creators["string_functions"] = create_string_functions

    has_json_extension = functions_exist(["json_array_length"])
    if not has_json_extension:

        def create_json_functions(conn):
            conn.create_function(
                "json_array_length", 1, py_json_array_length, deterministic=True
            )

        _registered_function_creators["json_functions"] = create_json_functions


def adapt_datetime(val: datetime) -> str:
    if not (val.tzinfo is timezone.utc or val.tzname() == "UTC"):
        try:
            val = val.astimezone(timezone.utc)
        except (OverflowError, ValueError, OSError):
            if val.year == MAXYEAR:
                val = datetime.max
            elif val.year == MINYEAR:
                val = datetime.min
            else:
                raise
    return val.replace(tzinfo=None).isoformat(" ")


def convert_datetime(val: bytes) -> datetime:
    return datetime.fromisoformat(val.decode()).replace(tzinfo=timezone.utc)


def path_parent(path):
    return func.rtrim(func.rtrim(path, func.replace(path, slash, empty_str)), slash)


def path_name(path):
    return func.ltrim(func.substr(path, func.length(path_parent(path)) + 1), slash)


def path_file_ext_length(path):
    name = path_name(path)
    expr = func.length(name) - func.length(
        func.rtrim(name, func.replace(name, dot, empty_str))
    )
    return case((func.instr(name, dot) == 0, 0), else_=expr)


def path_file_stem(path):
    return func.rtrim(
        func.substr(path, 1, func.length(path) - path_file_ext_length(path)), dot
    )


def path_file_ext(path):
    return func.substr(path, func.length(path) - path_file_ext_length(path) + 1)


def compile_path_parent(element, compiler, **kwargs):
    return compiler.process(path_parent(*element.clauses.clauses), **kwargs)


def compile_path_name(element, compiler, **kwargs):
    return compiler.process(path_name(*element.clauses.clauses), **kwargs)


def compile_path_file_stem(element, compiler, **kwargs):
    return compiler.process(path_file_stem(*element.clauses.clauses), **kwargs)


def compile_path_file_ext(element, compiler, **kwargs):
    return compiler.process(path_file_ext(*element.clauses.clauses), **kwargs)


def compile_cosine_distance(element, compiler, **kwargs):
    run_compiler_hook("cosine_distance")
    return f"cosine_distance({compiler.process(element.clauses, **kwargs)})"


def compile_euclidean_distance(element, compiler, **kwargs):
    run_compiler_hook("euclidean_distance")
    return f"euclidean_distance({compiler.process(element.clauses, **kwargs)})"


def py_json_array_length(arr):
    return len(ujson.loads(arr))


def compile_array_length(element, compiler, **kwargs):
    return compiler.process(func.json_array_length(*element.clauses.clauses), **kwargs)


def compile_string_length(element, compiler, **kwargs):
    return compiler.process(func.length(*element.clauses.clauses), **kwargs)


def compile_string_split(element, compiler, **kwargs):
    return compiler.process(func.split(*element.clauses.clauses), **kwargs)


def compile_greatest(element, compiler, **kwargs):
    """
    Compiles a sql function for `greatest(*args)` taking 1 or more args

    Compiles to:
      - `max(arg1, arg2...)` for 2 or more args
      - `arg1` for 1 arg

    sqlite's max() is a simple function when it has 2 or more
    arguments but operates as an aggregate function if given only a
    single argument
    See https://www.sqlite.org/lang_corefunc.html#max_scalar
    """
    args = element.clauses.clauses
    nargs = len(args)
    if nargs < 1:
        raise TypeError(
            f"conditional.greatest requires at least 1 argument ({nargs} found)"
        )
    if nargs == 1:
        expr = args[0]
    else:
        expr = func.max(*args)
    return compiler.process(expr, **kwargs)


def compile_least(element, compiler, **kwargs):
    """
    Compiles a sql function for `least(*args)` taking 1 or more args

    Compiles to:
      - `min(arg1, arg2...)` for 2 or more args
      - `arg1` for 1 arg

    sqlite's min() is a simple function when it has 2 or more
    arguments but operates as an aggregate function if given only a
    single argument
    See https://www.sqlite.org/lang_corefunc.html#min_scalar
    """
    args = element.clauses.clauses
    nargs = len(args)
    if nargs < 1:
        raise TypeError(
            f"conditional.least requires at least 1 argument ({nargs} found)"
        )
    if nargs == 1:
        expr = args[0]
    else:
        expr = func.min(*args)
    return compiler.process(expr, **kwargs)


def compile_values(element, compiler, **kwargs):
    return base_values_compiler(lambda i: f"column{i}", element, compiler, **kwargs)


def compile_rand(element, compiler, **kwargs):
    return compiler.process(func.random(), **kwargs)



import logging
import re

import PyPDF2
from pybtex.database.input import bibtex

from dvcx.query import DatasetRow, Object, udf
from dvcx.sql.types import Int, String

logger = logging.getLogger("dvcx")


def sanitize_utf8(original):
    """Clean up invalid UTF characters that may cause failures when writing to a DB."""
    return (
        original.encode("utf-8", "ignore").decode("utf-8", "ignore").replace("\x00", "")
    )


def sanitize_ascii(original):
    """More restrictive than sanitize_utf8 when we want just basic keyboard chars.
    This also puts all the text on a single line by eliminating all line breaks.
    If we expect a large amount of non-English text, then it makes sense to first use
    the Unidecode package.
    """
    sanitized = (
        original.encode("ascii", "ignore").decode("ascii", "ignore").replace("\x00", "")
    )
    sanitized = " ".join(sanitized.splitlines())
    sanitized = re.sub(r"[^\x20-\x7e]", r"", sanitized)
    # These should always pass after the previous conversions
    assert isinstance(sanitized, str)
    assert sanitized.isascii()
    assert sanitized.isprintable()
    return sanitized


def compress_whitespace(original):
    """PDF parsers often result in some weird whitespace."""
    return " ".join(original.split())


def load_pdf_pages(raw):
    raw_name = raw.info()["name"]

    pages = []
    try:
        pdf_reader = PyPDF2.PdfReader(raw)
        pages = [sanitize_utf8(page.extract_text()) for page in pdf_reader.pages]
    except Exception as error:  # noqa: BLE001
        error_type = type(error).__name__
        logger.warning(
            "A pdf reader error occurred in %s: %s - %s",
            raw_name,
            error_type,
            error,
        )
    return pages


@udf(
    params=(Object(load_pdf_pages), *tuple(DatasetRow.schema.keys())),
    output={**DatasetRow.schema, "n_page": Int, "total_pages": Int, "page": String},
)
def pdf_pages(pages, *args):
    record = dict(zip(DatasetRow.schema.keys(), args))
    del record["random"]  # random will be populated automatically
    record["is_latest"] = record["is_latest"] > 0  # needs to be a bool

    total_pages = len(pages)
    for n_page, page in enumerate(pages):
        page = sanitize_ascii(page)
        page = compress_whitespace(page)
        yield (*DatasetRow.create(**record), n_page, total_pages, page)


def load_bibtex(raw):
    bibtex_str = raw.read().decode("utf-8")
    parser = bibtex.Parser()
    return parser.parse_string(bibtex_str)



import os
import posixpath
from datetime import datetime, timezone
from pathlib import Path
from typing import TYPE_CHECKING, Any
from urllib.parse import urlparse

from fsspec.implementations.local import LocalFileSystem

from dvcx.node import Entry
from dvcx.storage import StorageURI

from .fsspec import Client

if TYPE_CHECKING:
    from dvcx.data_storage import AbstractMetastore


class FileClient(Client):
    FS_CLASS = LocalFileSystem
    PREFIX = "file://"
    protocol = "file"

    def __init__(
        self, name: str, fs: LocalFileSystem, cache, use_symlinks: bool = False
    ) -> None:
        super().__init__(name, fs, cache)
        self.use_symlinks = use_symlinks

    def url(self, path: str, expires: int = 3600, **kwargs) -> str:
        raise TypeError("Signed urls are not implemented for local file system")

    @property
    def uri(self) -> StorageURI:
        """
        This returns root of FS as uri, e.g
            Linux & Mac : file:///
            Windows: file:///C:/
        """
        return StorageURI(Path(self.name).as_uri())

    @staticmethod
    def root_dir() -> str:
        """
        Returns file system root path.
        Linux &  MacOS: /
        Windows: C:/
        """
        return Path.cwd().anchor.replace(os.sep, posixpath.sep)

    @staticmethod
    def root_path() -> Path:
        return Path(FileClient.root_dir())

    @classmethod
    def ls_buckets(cls, **kwargs):
        return []

    @classmethod
    def path_to_uri(cls, path: str) -> str:
        """
        Resolving path, that can be absolute or relative, to file URI which
        starts with file:/// prefix
        In unix like systems we support home shortcut as well.
        Examples:
            ./animals -> file:///home/user/working_dir/animals
            ~/animals -> file:///home/user/animals
            /home/user/animals -> file:///home/user/animals
            C:\\windows\animals -> file:///C:/windows/animals
        """
        return Path(path).expanduser().absolute().resolve().as_uri()

    @classmethod
    def split_url(cls, url: str) -> tuple[str, str]:
        """
        Splits url into two components:
            1. root of the FS which will later on become the name of the storage
            2. path which will later on become partial path
        Note that URL needs to be have file:/// protocol.
        Examples:
            file:///tmp/dir -> / + tmp/dir
            file:///c:/windows/files -> c:/ + windows/files
        """
        parsed = urlparse(url)
        if parsed.scheme == "file":
            scheme, rest = url.split(":", 1)
            uri = f"{scheme.lower()}:{rest}"
        else:
            uri = cls.path_to_uri(url)

        return cls.root_dir(), uri.removeprefix(cls.root_path().as_uri())

    @classmethod
    def from_name(
        cls, name: str, metastore: "AbstractMetastore", cache, kwargs
    ) -> "FileClient":
        use_symlinks = kwargs.pop("use_symlinks", False)
        return cls(name, cls.create_fs(**kwargs), cache, use_symlinks=use_symlinks)

    @classmethod
    def from_source(
        cls,
        uri: str,
        cache,
        use_symlinks: bool = False,
        **kwargs,
    ) -> "FileClient":
        fs = cls.create_fs(**kwargs)
        return cls(
            fs._strip_protocol(uri),
            cls.create_fs(**kwargs),
            cache,
            use_symlinks=use_symlinks,
        )

    async def get_current_etag(self, uid) -> str:
        info = self.fs.info(self.get_full_path(uid.path))
        return self.convert_info(info, "").etag

    async def get_size(self, path: str) -> int:
        return self.fs.size(path)

    async def get_file(self, lpath, rpath, callback):
        return self.fs.get_file(lpath, rpath, callback=callback)

    async def ls_dir(self, path):
        return self.fs.ls(path, detail=True)

    def rel_path(self, path):
        return posixpath.relpath(path, self.name)

    def get_full_path(self, rel_path):
        full_path = Path(self.name, rel_path).as_uri()
        if rel_path.endswith("/") or not rel_path:
            full_path += "/"
        return full_path

    def convert_info(self, v: dict[str, Any], parent: str) -> Entry:
        name = posixpath.basename(v["name"])
        return Entry.from_file(
            parent=parent,
            name=name,
            etag=v["mtime"].hex(),
            is_latest=True,
            last_modified=datetime.fromtimestamp(v["mtime"], timezone.utc),
            size=v.get("size", ""),
        )

    def fetch_nodes(
        self,
        nodes,
        shared_progress_bar=None,
    ) -> None:
        if not self.use_symlinks:
            super().fetch_nodes(nodes, shared_progress_bar)

    def do_instantiate_object(self, uid, dst):
        if self.use_symlinks:
            os.symlink(Path(self.name, uid.path), dst)
        else:
            super().do_instantiate_object(uid, dst)



from sqlalchemy.sql.elements import literal
from sqlalchemy.sql.expression import column

from . import functions
from .default import setup as default_setup
from .selectable import select, values

__all__ = [
    "column",
    "literal",
    "select",
    "values",
    "functions",
]

default_setup()



import json
import logging
from collections.abc import Iterable, Iterator
from datetime import datetime, timedelta, timezone
from struct import unpack
from typing import (
    Any,
    Generic,
    Optional,
    TypeVar,
)

from dvcx.dataset import DatasetStats
from dvcx.utils import retry_with_backoff

T = TypeVar("T")
LsData = Optional[list[dict[str, Any]]]
DatasetInfoData = Optional[dict[str, Any]]
DatasetStatsData = Optional[DatasetStats]
DatasetRowsData = Optional[Iterable[dict[str, Any]]]
DatasetExportStatus = Optional[dict[str, Any]]
DatasetExportSignedUrls = Optional[list[str]]


logger = logging.getLogger("dvcx")

DATASET_ROWS_CHUNK_SIZE = 8192


def _is_server_error(status_code: int) -> bool:
    return str(status_code).startswith("5")


def _parse_dates(obj: dict, date_fields: list[str]):
    """
    Function that converts string ISO dates to datetime.datetime instances in object
    """
    for date_field in date_fields:
        if obj.get(date_field):
            obj[date_field] = datetime.fromisoformat(obj[date_field])


class Response(Generic[T]):
    def __init__(self, data: T, ok: bool, message: str) -> None:
        self.data = data
        self.ok = ok
        self.message = message

    def __repr__(self):
        return (
            f"{self.__class__.__name__}(ok={self.ok}, data={self.data}"
            f", message={self.message})"
        )


class StudioClient:
    def __init__(
        self, url: str, username: str, token: str, timeout: float = 3600.0
    ) -> None:
        self._check_dependencies()
        self.url = url.rstrip("/")
        self.username = username
        self.token = token
        self.timeout = timeout

    def _check_dependencies(self) -> None:
        try:
            import msgpack  # noqa: F401
            import requests  # noqa: F401
        except ImportError as exc:
            raise Exception(
                f"Missing dependency: {exc.name}\n"
                "To install run:\n"
                "\tpip install 'dvcx[remote]'"
            ) from None

    def _send_request_msgpack(self, route: str, data: dict[str, Any]) -> Response[Any]:
        import msgpack
        import requests

        response = requests.post(
            f"{self.url}/{route}",
            json={**data, "team_name": self.username},
            headers={
                "Content-Type": "application/json",
                "Authorization": f"token {self.token}",
            },
            timeout=self.timeout,
        )
        ok = response.ok
        content = msgpack.unpackb(response.content, ext_hook=self._unpacker_hook)
        response_data = content.get("data")
        if ok and response_data is None:
            message = "Indexing in progress"
        else:
            message = content.get("message", "")
        return Response(response_data, ok, message)

    @retry_with_backoff(retries=5)
    def _send_request(self, route: str, data: dict[str, Any]) -> Response[Any]:
        """
        Function that communicate Studio API.
        It will raise an exception, and try to retry, if 5xx status code is
        returned, or if ConnectionError or Timeout exceptions are thrown from
        requests lib
        """
        import requests

        response = requests.post(
            f"{self.url}/{route}",
            json={**data, "team_name": self.username},
            headers={
                "Content-Type": "application/json",
                "Authorization": f"token {self.token}",
            },
            timeout=self.timeout,
        )
        try:
            response.raise_for_status()
        except requests.exceptions.HTTPError:
            if _is_server_error(response.status_code):
                # going to retry
                raise

        ok = response.ok
        try:
            data = json.loads(response.content.decode("utf-8"))
        except json.decoder.JSONDecodeError:
            data = {}

        if not ok:
            logger.error(
                "Got bad response from Studio, content is %s",
                response.content.decode("utf-8"),
            )
            if response.status_code == 403:
                message = "Not authorized"
            else:
                message = data.get("message", "")
        else:
            message = ""

        return Response(data, ok, message)

    @staticmethod
    def _unpacker_hook(code, data):
        import msgpack

        if code == 42:  # for parsing datetime objects
            has_timezone = False
            timezone_offset = None
            if len(data) == 8:
                # we send only timestamp without timezone if data is 8 bytes
                values = unpack("!d", data)
            else:
                has_timezone = True
                values = unpack("!dl", data)

            timestamp = values[0]
            if has_timezone:
                timezone_offset = values[1]
                return datetime.fromtimestamp(
                    timestamp, timezone(timedelta(seconds=timezone_offset))
                )
            return datetime.fromtimestamp(timestamp)  # noqa: DTZ006

        return msgpack.ExtType(code, data)

    def ls(self, paths: Iterable[str]) -> Iterator[tuple[str, Response[LsData]]]:
        # TODO: change LsData (response.data value) to be list of lists
        # to handle cases where a path will be expanded (i.e. globs)
        response: Response[LsData]
        for path in paths:
            response = self._send_request_msgpack("ls", {"source": path})
            yield path, response

    def dataset_info(self, name: str) -> Response[DatasetInfoData]:
        def _parse_dataset_info(dataset_info):
            _parse_dates(dataset_info, ["created_at", "finished_at"])
            for version in dataset_info.get("versions"):
                _parse_dates(version, ["created_at"])

            return dataset_info

        response = self._send_request("dataset-info", {"dataset_name": name})
        if response.ok:
            response.data = _parse_dataset_info(response.data)
        return response

    def dataset_rows_chunk(
        self, name: str, version: int, offset: int
    ) -> Response[DatasetRowsData]:
        def _parse_row(row):
            row["id"] = int(row["id"])
            return row

        req_data = {"dataset_name": name, "dataset_version": version}
        response = self._send_request_msgpack(
            "dataset-rows",
            {**req_data, "offset": offset, "limit": DATASET_ROWS_CHUNK_SIZE},
        )
        if response.ok:
            response.data = [_parse_row(r) for r in response.data]

        return response

    def dataset_stats(self, name: str, version: int) -> Response[DatasetStatsData]:
        response = self._send_request(
            "dataset-stats", {"dataset_name": name, "dataset_version": version}
        )
        if response.ok:
            response.data = DatasetStats(**response.data)
        return response

    def export_dataset_table(
        self, name: str, version: int
    ) -> Response[DatasetExportSignedUrls]:
        return self._send_request(
            "dataset-export", {"dataset_name": name, "dataset_version": version}
        )

    def dataset_export_status(
        self, name: str, version: int
    ) -> Response[DatasetExportStatus]:
        return self._send_request(
            "dataset-export-status", {"dataset_name": name, "dataset_version": version}
        )



import os
import os.path
import signal
import subprocess
import sys
from io import TextIOWrapper
from textwrap import dedent
from threading import Thread
from typing import Callable

import pytest

tests_dir = os.path.dirname(os.path.abspath(__file__))

python_exc = sys.executable or "python3"


E2E_STEP_TIMEOUT_SEC = 30


E2E_STEPS = (
    {
        "command": (
            "dvcx",
            "find",
            "--anon",
            "--name",
            "cat.1.*",
            "gcs://dvcx-datalakes/dogs-and-cats/",
        ),
        "expected": dedent(
            """
            gcs://dvcx-datalakes/dogs-and-cats/cat.1.jpg
            gcs://dvcx-datalakes/dogs-and-cats/cat.1.json
            """
        ),
        "listing": True,
    },
    {
        "command": (
            python_exc,
            os.path.join(tests_dir, "scripts", "name_len_normal.py"),
        ),
        "expected": dedent(
            """
            [('cat.1.jpg', 9), ('cat.1.json', -1)]
            """
        ),
    },
    {
        "command": (
            python_exc,
            os.path.join(tests_dir, "scripts", "name_len_slow.py"),
        ),
        "interrupt_after": "UDF Processing Started",
        "expected_in_stderr": "KeyboardInterrupt",
        "expected_not_in_stderr": "semaphore",
    },
    {
        "command": ("dvcx", "gc"),
        "expected": "Nothing to clean up.\n",
    },
)


def watch_process_thread(
    stream: TextIOWrapper, output_lines: list[str], watch_value: str, callback: Callable
) -> None:
    """
    Watches either the stdout or stderr stream from a given process,
    reads the output into output_lines, and watches for the given watch_value,
    then calls callback once found.
    """
    while (line := stream.readline()) != "":
        line = line.strip()
        output_lines.append(line)
        if watch_value in line:
            callback()


def communicate_and_interrupt_process(
    process: subprocess.Popen, interrupt_after: str
) -> tuple[str, str]:
    def interrupt_step() -> None:
        if sys.platform == "win32":
            # Windows has a different mechanism of sending a Ctrl-C event.
            process.send_signal(signal.CTRL_C_EVENT)
        else:
            process.send_signal(signal.SIGINT)

    stdout_lines: list[str] = []
    stderr_lines: list[str] = []
    watch_threads = (
        Thread(
            target=watch_process_thread,
            name="Test-Query-E2E-Interrupt-stdout",
            daemon=True,
            args=[process.stdout, stdout_lines, interrupt_after, interrupt_step],
        ),
        Thread(
            target=watch_process_thread,
            name="Test-Query-E2E-Interrupt-stderr",
            daemon=True,
            args=[process.stderr, stderr_lines, interrupt_after, interrupt_step],
        ),
    )
    for t in watch_threads:
        t.start()
    process.wait(timeout=E2E_STEP_TIMEOUT_SEC)
    return "\n".join(stdout_lines), "\n".join(stderr_lines)


def run_step(step):
    """Run an end-to-end query test step with a command and expected output."""
    command = step["command"]
    # Note that a process.returncode of -2 is the same as the shell returncode of 130
    # (canceled by KeyboardInterrupt)
    interrupt_exit_code = -2
    if sys.platform == "win32":
        # Windows has a different mechanism of creating a process group.
        popen_args = {"creationflags": subprocess.CREATE_NEW_PROCESS_GROUP}
        # This is STATUS_CONTROL_C_EXIT which is equivalent to 0xC000013A
        interrupt_exit_code = 3221225786
    else:
        popen_args = {"start_new_session": True}
    process = subprocess.Popen(
        command,
        shell=False,  # noqa: S603
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        encoding="utf-8",
        **popen_args,
    )
    interrupt_after = step.get("interrupt_after")
    if interrupt_after:
        stdout, stderr = communicate_and_interrupt_process(process, interrupt_after)
    else:
        stdout, stderr = process.communicate(timeout=E2E_STEP_TIMEOUT_SEC)

    if interrupt_after:
        if process.returncode not in (interrupt_exit_code, 1):
            print(f"Process stdout: {stdout}")
            print(f"Process stderr: {stderr}")
            raise RuntimeError(
                "Query script failed to interrupt correctly: "
                f"{process.returncode} Command: {command}"
            )
    elif process.returncode != 0:
        print(f"Process stdout: {stdout}")
        print(f"Process stderr: {stderr}")
        raise RuntimeError(
            "Query script failed with exit code: "
            f"{process.returncode} Command: {command}"
        )

    if step.get("sort_expected_lines"):
        assert sorted(stdout.split("\n")) == sorted(
            step["expected"].lstrip("\n").split("\n")
        )
    elif step.get("expected_in_stderr"):
        assert step["expected_in_stderr"] in stderr
        if step.get("expected_not_in_stderr"):
            assert step["expected_not_in_stderr"] not in stderr
    else:
        assert stdout == step["expected"].lstrip("\n")

    if step.get("listing"):
        assert "Listing" in stderr
    else:
        assert "Listing" not in stderr


@pytest.mark.e2e
def test_query_e2e(tmp_dir):
    """End-to-end CLI Query Test"""
    for step in E2E_STEPS:
        run_step(step)



from sqlalchemy.ext.compiler import compiles
from sqlalchemy.sql import expression, selectable


class Values(selectable.Values):
    def __init__(self, data, columns=None, **kwargs):
        if columns is None:
            num_columns = len(data[0])
            columns = [expression.column(f"c{i}") for i in range(1, num_columns + 1)]
        else:
            columns = [
                expression.column(c) if isinstance(c, str) else c for c in columns
            ]
        super().__init__(*columns, **kwargs)
        self._data += tuple(data)


def values(data, columns=None, **kwargs) -> Values:
    return Values(data, columns=columns, **kwargs)


def process_column_expressions(columns):
    return [expression.column(c) if isinstance(c, str) else c for c in columns]


def select(*columns, **kwargs) -> "expression.Select":
    columns = process_column_expressions(columns)
    return expression.select(*columns, **kwargs)


def base_values_compiler(column_name_func, element, compiler, **kwargs):
    columns = element.columns
    base_values = expression.values(*columns).data(element._data)
    col_expressions = [
        expression.column(column_name_func(i)).label(c.name)
        for i, c in enumerate(columns, 1)
    ]
    expr = (
        expression.select(*col_expressions)
        .select_from(base_values)
        .subquery(element.name)
    )
    return compiler.process(expr, **kwargs)


def compile_values(element, compiler, **kwargs):
    return base_values_compiler(lambda i: f"c{i}", element, compiler, **kwargs)


compiles(Values, "default")(compile_values)



import inspect
from typing import TYPE_CHECKING, Any, Callable, Optional

from dvcx.lib.file import TextFile
from dvcx.lib.reader import FeatureReader

if TYPE_CHECKING:
    from dvcx.lib.feature_types import FeatureLike


def convert_text(
    text: str,
    tokenizer: Optional[Callable] = None,
    tokenizer_kwargs: Optional[dict[str, Any]] = None,
    open_clip_model: Optional[Any] = None,
):
    """
    Tokenize and otherwise transform text.

    Args:
        text (str): Text to convert.
        tokenizer (Callable): Tokenizer to use to tokenize objects.
        tokenizer_kwargs (dict): Additional kwargs to pass when calling tokenizer.
        open_clip_model (Any): Encode text using model from open_clip library.
    """
    if open_clip_model:
        method_name = "encode_text"
        if not (
            hasattr(open_clip_model, method_name)
            and inspect.ismethod(getattr(open_clip_model, method_name))
        ):
            raise ValueError(
                f"TextColumn error: 'model' doesn't support '{method_name}()'"
            )

    if not tokenizer:
        return text

    if tokenizer_kwargs:
        res = tokenizer([text], **tokenizer_kwargs)
    else:
        res = tokenizer([text])
    from transformers.tokenization_utils_base import PreTrainedTokenizerBase

    tokens = (
        res.input_ids[0] if isinstance(tokenizer, PreTrainedTokenizerBase) else res[0]
    )
    if not open_clip_model:
        return tokens

    return open_clip_model.encode_text(tokens.unsqueeze(0))


class TextReader(FeatureReader):
    def __init__(
        self,
        fr_class: "FeatureLike" = TextFile,
        tokenizer: Optional[Callable] = None,
        tokenizer_kwargs: Optional[dict[str, Any]] = None,
        open_clip_model: Optional[Any] = None,
    ):
        """
        Read and optionally transform a text column.

        All kwargs are passed to `convert_text()`.
        """
        self.tokenizer = tokenizer
        self.tokenizer_kwargs = tokenizer_kwargs
        self.open_clip_model = open_clip_model
        super().__init__(fr_class)

    def __call__(self, value: str):
        return convert_text(
            value,
            tokenizer=self.tokenizer,
            tokenizer_kwargs=self.tokenizer_kwargs,
            open_clip_model=self.open_clip_model,
        )
